"""empty message

Revision ID: roc_dingo_0020_rm_biacurrtablehfr
Revises: roc_dingo_0019_modify_views_per_date
Create Date: 2021-12-17 10:13:35.507837

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

from poppy.pop.alembic.helpers import \
    create_table, execute, user, drop_table

# revision identifiers, used by Alembic.
revision = 'roc_dingo_0020_rm_biacurrtablehfr'
down_revision = 'roc_dingo_0019_modify_views_per_date'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    drop_table('bia_current_log',
               schema='pipeline')
    drop_table('bia_sweep_table',
               schema='pipeline')
    drop_table('hfr_freq_list',
               schema='pipeline')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    create_table('bia_sweep_table',
    sa.Column('id', sa.BIGINT(), nullable=False),
    sa.Column('tc_name', sa.String(length=256), nullable=False),
    sa.Column('utc_time', postgresql.TIMESTAMP(), nullable=False),
    sa.Column('sweep_eeprom', sa.BOOLEAN(), nullable=True),
    sa.Column('sweep_step_cur', sa.String(), nullable=True),
    sa.Column('insert_time', postgresql.TIMESTAMP(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('utc_time', 'tc_name'),
    schema='pipeline'
    )
    create_table('hfr_freq_list',
    sa.Column('id', sa.BIGINT(), nullable=False),
    sa.Column('utc_time', postgresql.TIMESTAMP(), nullable=False),
    sa.Column('tc_name', sa.String(length=256), nullable=False),
    sa.Column('freq_list', sa.String(), nullable=False),
    sa.Column('freq_nr', sa.INTEGER(), nullable=False),
    sa.Column('insert_time', postgresql.TIMESTAMP(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('utc_time', 'tc_name'),
    schema='pipeline'
    )
    create_table('bia_current_log',
                 sa.Column('id', sa.BIGINT(), nullable=False),
                 sa.Column('tc_name', sa.String(length=256), nullable=False),
                 sa.Column('utc_time', postgresql.TIMESTAMP(), nullable=False),
                 sa.Column('bia_current', sa.FLOAT(), nullable=False),
                 sa.Column('insert_time', postgresql.TIMESTAMP(),
                           nullable=True),
                 sa.PrimaryKeyConstraint('id'),
                 sa.UniqueConstraint('utc_time', 'tc_name'),
                 schema='pipeline'
                 )

    # added by hand
    # fix table permissions
    grant_user_seq = (
        """GRANT USAGE, SELECT
        ON ALL SEQUENCES IN SCHEMA pipeline TO {0}""".format(user)
    )
    execute(grant_user_seq)
    # ### end Alembic commands ###



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to synchronize the file tree with the ROC database."""

import urllib.request
import urllib.parse
from urllib.error import HTTPError
import json

from datetime import timedelta

from sqlalchemy import and_
from sqlalchemy.sql import func
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import load_only

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task
from poppy.core.target import BaseTarget

from roc.dingo.constants import PIPELINE_DATABASE
from roc.dingo.models.file import FileLog


__all__ = ['CheckSOARDataAvailability']


class CheckSOARDataAvailability(Task):
    """
    For each files flagged as delivered in DB, search SOAR if it available
    if yes : set is_archived
    """

    plugin_name = 'roc.dingo'
    name = 'check_soar'

    def add_targets(self):
        logger.debug('CheckSOARDataAvailability() : add_targets')

        # number of archived files
        self.add_output(target_class=BaseTarget,
                        many=True,
                        identifier='store_archived')

    def setup_inputs(self):
        """
        Setup task inputs.

        :param task:
        :return:
        """

        logger.debug('CheckSOARDataAvailability() : setup_inputs')

        # get the TAP URL
        self.soar_tap_url = self.pipeline.get('url', args=True)

    def run(self):

        logger.debug('CheckSOARDataAvailability() : run')

        # Initialize inputs
        self.setup_inputs()

        # Initialize counters
        cpt_archived = 0          # files really stored as is_archived
        cpt_to_archive = 0        # files to be stored as is_archived
        cpt_files_to_archive = 0  # files delivered but not archived yet
        ids_archived = []
        min_start_time = None
        max_start_time = None

        # Get the database connection if needed
        if not hasattr(self, 'session'):
            self.session = Connector.manager[PIPELINE_DATABASE].session

        # Find entries in file_log with is_delivered and not is_archived
        is_not_archived_filter = FileLog.is_archived == False   # noqa: E712
        is_delivered_filter = FileLog.is_delivered == True   # noqa: E712
        fields = ['id', 'basename', 'public_filename']

        try:
            query = self.session.query(FileLog).\
                options(load_only(*fields)).\
                filter(and_(is_not_archived_filter, is_delivered_filter))
            results = query.all()
        except Exception:
            logger.error('Query failed :')
            logger.error(
                str(query.statement.compile(
                    dialect=postgresql.dialect(),
                    compile_kwargs={'literal_binds': True})))
            results = []

        cpt_files_to_archive = len(results)
        logger.info(f'{cpt_files_to_archive} files to store as is_archived')

        # Find min anx max times for these files
        # This will be use to filter the TAP query, in order to limit
        # the data volume to retreive
        try:
            query_min = self.session.query(
                func.min(FileLog.start_time), func.max(FileLog.end_time)).\
                filter(and_(is_not_archived_filter, is_delivered_filter))
            min_start_time = query_min.one()[0]
            max_start_time = query_min.one()[1]
        except Exception:
            logger.error('Query failed :')
            logger.error(
                str(query_min.statement.compile(
                    dialect=postgresql.dialect(),
                    compile_kwargs={'literal_binds': True})))

        if min_start_time is not None:
            min_start_time = min_start_time - timedelta(days=1)
            min_start_time = min_start_time.strftime('%Y-%m-%d')
            logger.debug(f'Filtering on begin_time > {min_start_time}')

        if max_start_time is not None:
            max_start_time = max_start_time + timedelta(days=1)
            max_start_time = max_start_time.strftime('%Y-%m-%d')
            logger.debug(f'Filtering on begin_time < {max_start_time}')

        # Build URL of the query
        select_query = ['SELECT', 'file_name', 'FROM', 'v_public_files']

        where_query = ['WHERE', 'instrument', '=', '%27RPW%27']
        if min_start_time is not None:
            gt = urllib.parse.quote('>')
            where_query += ['AND', 'begin_time', gt, f'%27{min_start_time}%27']
        if max_start_time is not None:
            lt = urllib.parse.quote('<')
            where_query += ['AND', 'begin_time', lt, f'%27{max_start_time}%27']

        order_query = ['ORDER', 'BY', 'archived_on', 'ASC']

        tap_query = '+'.join(select_query + where_query + order_query)
        query_params = '&'.join([
            'REQUEST=doQuery',
            'LANG=ADQL',
            'FORMAT=json',
            f'QUERY={tap_query}'
        ])
        query_url = f'{self.soar_tap_url}/sync?{query_params}'
        logger.debug(query_url)

        # Run TAP query
        data = None
        try:
            logger.info(f'Querying --> {query_url}')
            data = CheckSOARDataAvailability.soar_run_query(query_url)
        except HTTPError as e:
            logger.error(f'SOAR TAP query has failed ({query_url})')
            logger.error(e)
            logger.exception(e.msg)

        # re-organise data
        archived_data = [d[0] for d in data['data']]

        for r in results:
            if r.public_filename in archived_data:
                logger.info(f'{r.basename} -> is_archived')
                ids_archived.append(r.id)
            else:
                logger.warning(f'{r.basename} not archived')

        cpt_to_archive = len(ids_archived)

        # update DB
        try:
            id_filter = FileLog.id.in_(ids_archived)
            query = self.session.query(FileLog).filter(id_filter)

            item = {}
            item['is_archived'] = True
            cpt_archived = query.update(item, synchronize_session=False)
        except Exception as e:
            logger.error(f'is_archived update has failed: \n {e}')
            logger.error(
                str(query.statement.compile(
                    dialect=postgresql.dialect(),
                    compile_kwargs={'literal_binds': True})))

        if cpt_archived != cpt_to_archive:
            # this should not happen
            logger.warning(f'[{cpt_archived}/{cpt_to_archive}] files '
                           'stored as is_archived')
        else:
            logger.info(f'{cpt_archived} files stored as is_archived')

        # number of archived files
        self.outputs['store_archived'].data = {
            'files_to_archive': cpt_files_to_archive,
            'files_archived': cpt_archived,
        }

    @staticmethod
    def soar_run_query(query: str) -> dict:
        """
        Runs a ADQL query on SOAR TAP server and return the data table.

        :param query: ADQL query_adql full URL string
        :return: a dictionary containing the query results
        """
        with urllib.request.urlopen(query) as url:
            data = json.loads(url.read().decode())
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" Tasks for handling files processed by DINGO plugin."""

import os
from pathlib import Path

from poppy.core.logger import logger
from poppy.core.task import Task

__all__ = ['MoveProcessedFiles', 'MoveFailedFiles']

from roc.dingo.tools import safe_move


class MoveProcessedFiles(Task):
    """Task to move input list of well processed files in a given target directory."""
    plugin_name = 'roc.dingo'
    name = 'move_processed_files'

    def setup_inputs(self):

        # See if --no-move keyword is defined
        self.no_move = self.pipeline.get('no_move', default=False, args=True)

        # Get or create processed_files list from pipeline properties
        self.processed_files = self.pipeline.get(
            'processed_files', default=[], create=True)

        # Get or create processed_files_dir from pipeline properties
        self.processed_files_dir = self.pipeline.get(
            'processed_files_dir', default=[None], create=True)[0]

    def run(self):

        try:
            self.setup_inputs()
        except:
            logger.exception(
                'Initializing inputs for task MoveProcessedFiles has failed!')
            self.pipeline.exit()
            return

        if self.no_move:
            logger.debug(
                'Skip MoveProcessedFiles task')
            return

        if not self.processed_files:
            logger.debug(
                'Input list of processed files is empty: skip MoveProcessedFiles task')
            return
        else:

            if not self.processed_files_dir:
                logger.debug(
                    'processed_files_dir argument not passed, skip task')
                return

            # Create folder if it does not exist
            Path(self.processed_files_dir).mkdir(parents=True, exist_ok=True)

            for current_file in self.processed_files:
                logger.info(f'Moving {current_file} into {self.processed_files_dir}')
                if not safe_move(current_file, self.processed_files_dir):
                    logger.error(f'Cannot move {current_file} into {self.processed_files_dir}')


class MoveFailedFiles(Task):
    """Move failed files found
    into a target directory."""
    plugin_name = 'roc.dingo'
    name = 'move_failed_files'

    def setup_inputs(self):

        # See if --no-move keyword is defined
        self.no_move = self.pipeline.get('no_move', default=False, args=True)

        # Get or create failed_files list from pipeline properties
        self.failed_files = self.pipeline.get(
            'failed_files', default=[], create=True)

        # Get or create failed_files_dir list from pipeline properties
        self.failed_files_dir = self.pipeline.get(
            'failed_files_dir', default=[None], create=True)[0]

    def run(self):

        try:
            self.setup_inputs()
        except:
            logger.exception(
                'Initializing inputs for task MoveFailedFiles has failed!')
            self.pipeline.exit()
            return

        if self.no_move:
            logger.debug(
                'Skip current task MoveFailedFiles')
            return

        if not self.failed_files:
            logger.debug(
                'Input list of failed files is empty: skip MoveFailedFiles task')
            return
        else:

            if not self.failed_files_dir:
                logger.warning(f'There are failed files but failed_files_dir value is not defined, skip MoveFailedFiles task')
                return

            # Create folder if it does not exist
            Path(self.failed_files_dir).mkdir(parents=True, exist_ok=True)

            for current_file in self.failed_files:
                logger.info(f'Moving {current_file} into {self.failed_files_dir}')
                # If failed file does not exist, create an empty file
                if not os.path.isfile(current_file):
                    Path(current_file).touch(exist_ok=True)

                if not safe_move(current_file, self.failed_files_dir):
                    logger.error(f'Cannot move {current_file} into {self.failed_files_dir}')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Tests module for the roc.dingo plugin.
"""
import pytest
import os.path as osp
from roc.dingo.models.file import FileLog
from roc.dingo.tests.test_log_file import \
    DingoLogFileCommandTestCase, DingoLogFileTest


# Tests on roc.dingo.tasks.log_file methods
class TestDingoCheckSoar(DingoLogFileCommandTestCase):

    def setup_method(self, method):
        super().setup_method(method)

        # empty the database
        self.session.query(FileLog).delete()
        self.session.flush()

    def teardown_method(self, method):
        super().teardown_method(method)

        # close database
        self.session.close()

    @pytest.mark.parametrize(
        'root, public_root, delivered_root, nb_archived_files', [
            ('DATA-1', 'PUBLIC-1', 'DELIVERED-1', 72),
        ])
    def test_check_soar(self,
                        root,
                        public_root,
                        delivered_root,
                        nb_archived_files):
        """
        Insert dataset with logfile_to_rocdb, store delivered file,
        and after check the SOAR report
        """
        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)
        public_root = osp.join(data_test_path, public_root)
        delivered_root = osp.join(data_test_path, delivered_root)

        # --- initialize the Db with files from DATA-1 ---
        main_command = ['pop',
                        'dingo',
                        'logfile_to_rocdb',
                        '--root',
                        root
                        ]

        # --- run the command ---
        self.run_command(main_command)

        # --- insert delivered data ---
        main_command = ['pop',
                        'dingo',
                        'store_delivered',
                        '--public',
                        public_root,
                        '--delivered',
                        delivered_root
                        ]

        # --- run the command ---
        self.run_command(main_command)

        # --- check for SOAR data availability ---
        main_command = ['pop',
                        'dingo',
                        'check_soar'
                        ]

        # --- run the command ---
        self.run_command(main_command)

        # --- make assertions ---

        # Count the is_archived files
        results = self.session.query(FileLog).\
            filter(FileLog.is_archived == True).all()  # noqa: E712

        assert len(results) == nb_archived_files

        self.session.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Tests module for the roc.dingo plugin.
"""
import pytest
import os.path as osp
from roc.dingo.models.file import FileLog
from roc.dingo.tests.test_dingo import DingoTest
from roc.dingo.tests.test_log_file import \
    DingoLogFileCommandTestCase, DingoLogFileTest


# Tests on roc.dingo.tasks.log_file methods
class TestDingoLogFileUpdate(DingoLogFileCommandTestCase):

    def setup_method(self, method):
        super().setup_method(method)

        # empty the database
        self.session.query(FileLog).delete()
        self.session.flush()

    def teardown_method(self, method):
        super().teardown_method(method)

        # close database
        self.session.close()

    @pytest.mark.parametrize('root, nb_files, file_to_update', [
        ('DATA-1', 61, 'solo_L1_rpw-hfr-surv-cdag_20200819_V07.cdf'),
    ])
    def test_logfile_update(self, root, nb_files, file_to_update):
        """
        After data insertion, modify file_to_update and test its update
        """
        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)

        # --- initialize the command ---
        main_command = ['pop',
                        'dingo',
                        'logfile_to_rocdb',
                        '--root', root
                        ]

        # --- run the command ---
        self.run_command(main_command)

        results = self.session.query(FileLog).all()
        assert len(results) == nb_files

        # modify DB to mark a file to be updated
        # and set to null some fields

        db_file_before = DingoTest.get_db_values(
            self.session.query(FileLog).
            filter(FileLog.basename == file_to_update).one())

        db_file = {}
        db_file['to_update'] = True
        db_file['creation_time'] = '1970-01-01'
        db_file['url'] = ''
        db_file['dataset_id'] = ''
        db_file['version'] = '0'
        db_file['start_time'] = '1970-01-01'
        db_file['end_time'] = '1970-01-01'
        db_file['product'] = ''
        db_file['is_delivered'] = True
        db_file['is_archived'] = True
        db_file['public_filename'] = 'x'
        db_file['public_dirname'] = 'y'

        fields = db_file.keys()
        # these fields will not be changed by the update
        # because they are handled by another task
        not_changed_fields = ['is_delivered',
                              'is_archived',
                              'public_filename',
                              'public_dirname']

        self.session.query(FileLog).\
            filter(FileLog.basename == file_to_update).update(db_file)
        self.session.flush()

        # -- get the modified version
        db_file_modified = DingoTest.get_db_values(
            self.session.query(FileLog).
            filter(FileLog.basename == file_to_update).one())

        # --- make assertions ---
        for key in fields:
            assert db_file_modified[key] != db_file_before[key]

        # --- initialize the command ---
        main_command = ['pop',
                        'dingo',
                        'update_filelog',
                        '--root', root
                        ]

        # --- run the command ---
        self.run_command(main_command)

        # -- get the new version
        db_file_after = DingoTest.get_db_values(
            self.session.query(FileLog).
            filter(FileLog.basename == file_to_update).one())

        # --- make assertions ---
        for key in fields:
            if key in not_changed_fields:
                # these fields are not changed by update_filelog
                assert db_file_after[key] != db_file_before[key]
                assert db_file_after[key] == db_file_modified[key]
            else:
                # these fields have to get their value back
                assert db_file_after[key] == db_file_before[key]
                assert db_file_after[key] != db_file_modified[key]

        # -- also check that there was one file updated
        # wil be done after Outflow migration

        self.session.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to insert SOLO/RPW-related event data into the ROC database."""
import csv
import json
import os
import uuid
from datetime import datetime, timedelta

import numpy as np
import pandas as pd
from sqlalchemy import and_, or_, null

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task

from roc.dingo.models.data import EventLog, EfecsEvents, BiaSweepLog, SbmLog
from roc.dingo.constants import PIPELINE_DATABASE, TRYOUTS, TIME_WAIT_SEC, SQL_LIMIT, TF_PA_DPU_0038, TF_PA_DPU_0039, \
    IDB_SOURCE, EFECS_EVENT_LOG, TM_APID_EVENT_LOG, TC_PACKET_EVENT_LOG, HFR_FREQ_LIST_PACKETS, BIA_CURRENT_LOG_PACKETS, \
    BIA_SWEEP_TABLE_PACKETS, BIA_SWEEP_TABLE_NR, TIME_ISO_STRFORMAT, LFR_CALIB_EVENT_LOG
from roc.dingo.models.packet import TmLog, TcLog

from roc.dingo.tools import load_spice, query_db, gen_sql_filters, \
    get_columns, insert_in_db, sbm_qf_eng, \
    get_current_idb, raw_to_na, compute_hfr_list_freq

__all__ = ["EventToDb"]


class EventToDb(Task):
    """
    Insert SOLO/RPW-related event data
    (from database,  SOLO HK data)
    into the ROC database
    """
    plugin_name = 'roc.dingo'
    name = 'event_to_db'

    def add_targets(self):
        pass

    @Connector.if_connected(PIPELINE_DATABASE)
    def setup_inputs(self):

        # get the input files
        self.start_time = self.pipeline.get('start_time', default=[None])[0]
        self.end_time = self.pipeline.get('end_time', default=[None])[0]

        # Get SOLO SPICE kernels (SCLK and LSK)
        self.sclk_file = self.pipeline.get(
            'sclk', default=[None], create=True)[0]
        self.lsk_file = self.pipeline.get(
            'lsk', default=[None], create=True)[0]
        if not self.sclk_file or not self.lsk_file:
            raise FileNotFoundError(
                'Both sclk_file and lsk_file must be passed as inputs to run EventToDb-{self.job_sid}!')
        else:
            # Load SPICE kernels
            self.spice = load_spice(
                spice_kernels=[self.lsk_file, self.sclk_file])
        # Get tryouts from pipeline properties
        self.tryouts = self.pipeline.get(
            'tryouts', default=[TRYOUTS], create=True)[0]

        # Get wait from pipeline properties
        self.wait = self.pipeline.get(
            'wait', default=[TIME_WAIT_SEC], create=True)[0]

        # Retrieve --limit keyword value
        self.limit = self.pipeline.get('limit',
                                       default=[SQL_LIMIT],
                                       args=True)[0]

        # get a database session
        self.session = Connector.manager[PIPELINE_DATABASE].session

        # Get idb_version/idb_source from pipeline properties
        self.idb_source = self.pipeline.get(
            'idb_source',
            default=[IDB_SOURCE],
            create=True)[0]

        self.idb_version = self.pipeline.get(
            'idb_version',
            default=[None],
            create=True)[0]

        # If idb_version not passed (is None),
        # then try to get current working version from the database
        if self.idb_version is None:
            self.idb_version = get_current_idb(
                self.idb_source,
                self.session,
                tryouts=self.tryouts,
                wait=self.wait,
            )
        if self.idb_version is None:
            raise ValueError(f'idb_version argument cannot be defined!')

        # Initialize some class attributes
        self.bia_sweep_tables = None

    def run(self):
        # Define task job ID (long and short)
        self.job_id = str(uuid.uuid4())
        self.job_sid = self.job_id[:8]
        logger.info(f'Task EventToDb-{self.job_sid} is starting')
        try:
            self.setup_inputs()
        except:
            logger.exception(f'Initializing inputs has failed for task EventToDb-{self.job_sid}!')
            self.pipeline.exit()
            return

        # Inserting TM data into event_log table
        logger.info(f'Inserting pipeline.tm_log data into pipeline.event_log for task EventToDb-{self.job_sid}')
        filters = [TmLog.apid == current_tm
                   for current_tm in TM_APID_EVENT_LOG]
        filters = or_(*filters)
        filters = and_(filters, gen_sql_filters(TmLog,
                                                start_time=self.start_time,
                                                end_time=self.end_time))
        tm_inserted, tm_failed, tm_ignored = self.to_event_log(TmLog, self.tm_to_event,
                                                               filters=filters)

        # Inserting TC data into event_log table
        logger.info(f'Inserting pipeline.tc_log data into pipeline.event_log for task EventToDb-{self.job_sid}')
        filters = [TcLog.palisade_id == current_tc
                   for current_tc in TC_PACKET_EVENT_LOG]
        filters = or_(*filters)
        filters = and_(filters, gen_sql_filters(TcLog,
                                                start_time=self.start_time,
                                                end_time=self.end_time))
        tc_inserted, tc_failed, tc_ignored = self.to_event_log(TcLog, self.tc_to_event,
                                                               filters=filters)

        # Inserting E-FECS data into event_log table
        logger.info(f'Inserting pipeline.efecs_events data into pipeline.event_log for task EventToDb-{self.job_sid}')
        filters = [EfecsEvents.name == current_efecs
                   for current_efecs in EFECS_EVENT_LOG]
        filters = or_(*filters)
        filters = and_(filters, gen_sql_filters(EfecsEvents,
                                                start_time=self.start_time,
                                                end_time=self.end_time))
        efecs_inserted, efecs_failed, efecs_ignored = self.to_event_log(
            EfecsEvents, self.efecs_to_event,
            filters=filters)

        # Inserting SBM data into event_log table
        logger.info(f'Inserting pipeline.sbm_log data into pipeline.event_log for task EventToDb-{self.job_sid}')
        sbm_inserted, sbm_failed, sbm_ignored = self.to_event_log(
            SbmLog, self.sbm_to_event)

        # Inserting TC LFR calibration data into event_log table
        # For LFR calib, get all data since the beginning of the mission
        # (To be sure to not have incomplete calibrations)
        filters = [TcLog.palisade_id == current_tc
                   for current_tc in LFR_CALIB_EVENT_LOG]
        filters = or_(*filters)
        lfrcal_inserted, lfrcal_failed, lfrcal_ignored = self.to_event_log(
            TcLog, self.lfrcalib_to_event,
            filters=filters)

        # Inserting Bias sweep table data into event_log table
        # For sweep tables get all data since the beginning of the mission
        # (To be sure to have the full table loading history)
        filters = [TcLog.palisade_id == current_tc
                   for current_tc in BIA_SWEEP_TABLE_PACKETS]
        filters = or_(*filters)
        logger.info(f'Inserting bias sweep table data (in tc_log) into pipeline.event_log for task EventToDb-{self.job_sid}')
        sweeptable_inserted, sweeptable_failed, sweeptable_ignored = self.to_event_log(
            TcLog, self.biasweeptable_to_event,
            filters=filters)

        # Inserting Bias sweep data into event_log table
        # For Bias sweep get all data since the beginning of the mission
        # (To be sure to not have incomplete sweep)
        sweep_inserted, sweep_failed, sweep_ignored = self.to_event_log(
            BiaSweepLog, self.biasweep_to_event)

        inserted = sweep_inserted + sweeptable_inserted + \
            sbm_inserted + efecs_inserted + \
            tm_inserted + tc_inserted + lfrcal_inserted
        failed = sweep_failed + sweeptable_failed + \
            sbm_failed + efecs_failed + \
            tm_failed + tc_failed + lfrcal_failed
        ignored = sweep_ignored + sweeptable_ignored + \
            sbm_ignored + efecs_ignored + \
            tm_ignored + tc_ignored + lfrcal_ignored

        logger.info(f'{inserted} events inserted')
        if failed > 0:
            logger.error(f'{failed} events failed')
        if ignored > 0:
            logger.warning(f'{ignored} events ignored')

    def efecs_to_event(self, data):
        """
        Convert input data rows from pipeline.efecs_events table
        into entries to be saved in pipeline.event_log table.

        :param data: input efecs_events data rows as a pandas.DataFrame object
        :return: pandas.DataFrame object containing Efecs events to be inserted
        """

        # Sub-method to compute end_time from start_time and duration in seconds
        # (Used by pandas.DataFrame.apply() method to get end_time)
        def get_end_time(row):
            return row.utc_time + timedelta(
                seconds=int(row.attributes["duration"]))

        # Make sure to inserted allowed EFECS events
        efecs_to_insert = pd.DataFrame()
        n_data = data.shape[0]

        # Define output event_log columns values
        efecs_to_insert['label'] = data['name']
        efecs_to_insert['start_time'] = data['utc_time']
        efecs_to_insert['is_predictive'] = [True] * n_data

        # Define end_time (if any duration. Othewise end_time=start_time)
        efecs_to_insert['end_time'] = data.apply(
            lambda row: get_end_time(row)
            if 'attributes' in row else row.utc_time,
            axis=1)

        # Set description
        efecs_to_insert['description'] = data['attributes']

        return efecs_to_insert

    def sbm_to_event(self, data):
        """
        Convert input data rows from pipeline.sbm_log table
        into entries to be saved in pipeline.event_log table.

        :param data: input sbm_log data rows as a pandas.DataFrame object
        :return: a pandas.DataFrame object containing data to be inserted
        """
        # Sub-method to compute sbm event start_time/end_time
        def get_start_end(data):
            # extract SBM algo parameters
            # (stored as a strings)
            params = [int(val)
                      for key, val in data['sbm_algo_param'].items()]
            sbm_time = data["utc_time"]
            if data["sbm_type"] == 1 and len(params) == 3:
                sbm_dt1_sbm1 = params[0]
                sbm_dt2_sbm1 = params[1]
                sbm_dt3_sbm1 = params[2]
                # Compute SBM event start, end and occurred times
                # (see SSS or DAS User manual for details)
                if sbm_dt2_sbm1 < 2 * sbm_dt1_sbm1:
                    end_time = sbm_time + timedelta(
                        seconds=sbm_dt1_sbm1)
                    start_time = end_time - timedelta(
                        seconds=sbm_dt2_sbm1)
                elif sbm_dt2_sbm1 > 2 * sbm_dt1_sbm1:
                    end_time = sbm_time + \
                        timedelta(
                            seconds=int(sbm_dt1_sbm1 + sbm_dt3_sbm1))
                    start_time = end_time - timedelta(seconds=sbm_dt2_sbm1)
                else:
                    end_time = sbm_time + timedelta(
                        seconds=sbm_dt1_sbm1)
                    start_time = end_time - timedelta(
                        seconds=sbm_dt2_sbm1)

                # Set SBM1 quality factor transfer function
                sbm_qf_tf = TF_PA_DPU_0038

            elif data["sbm_type"] == 2 and len(params) == 10:
                sbm_dt = params[0]
                # Compute start_time / end_time
                start_time = sbm_time - timedelta(
                    seconds=(sbm_dt / 2 + 1))
                end_time = sbm_time + timedelta(
                    seconds=(sbm_dt / 2 + 1))

                # Set SBM1 quality factor transfer function
                sbm_qf_tf = TF_PA_DPU_0039
            else:
                logger.error(f'Entry {data} is badly formatted, please check!')
                return None, None, None

            # Compute engineering value of quality factor
            sbm_qf = sbm_qf_eng(float(data['sbm_qf']), sbm_qf_tf,
                                idb_source=self.idb_source,
                                idb_version=self.idb_version)

            # store sbm parameters as a JSON format string
            description = {key: val
                           for key, val in data.items()
                           if key in ['cuc_time', 'selected', 'status']
                           }
            # Add Quality factor
            description['sbm_qf'] = sbm_qf

            # And SBM algo parameters
            for key, val in data['sbm_algo_param'].items():
                description[key] = int(val)

            return start_time, end_time, description

        # initialize output
        sbm_to_insert = pd.DataFrame()

        # Create label and is_predictive
        sbm_to_insert['label'] = data["sbm_type"].apply(
            lambda row: f'SBM{row} EVENT')
        sbm_to_insert['is_predictive'] = data['utc_time_is_predictive']

        # Define start_time, end_time and description
        sbm_to_insert['start_time'], \
            sbm_to_insert['end_time'], sbm_to_insert['description'] = zip(*data.apply(
                lambda row: get_start_end(row),
                axis=1))

        return sbm_to_insert

    def biasweeptable_to_event(self, data):
        """
        Extract BIAS sweep table values from input TC data

        :param data: TC packet data (as a pandas.DataFrame object)
        :return: pandas.DataFrame object containing bias sweep tables to insert
        """

        def to_na(sweep_first_idx, sweep_step_nr, sweep_step_cur,
                  sweep_table=None):

            if sweep_table is None:
                sweep_table = np.empty(BIA_SWEEP_TABLE_NR, dtype=np.float32)
                sweep_table[:] = np.nan
            # Update the sweep table with new current values
            sweep_table[sweep_first_idx: sweep_first_idx + sweep_step_nr] = raw_to_na(
                sweep_step_cur, idb_version=self.idb_version, idb_source=self.idb_source)

            # Return current table
            return sweep_table

        # Initialize output
        output_data = []

        # Initialize loop variables
        # We assume the sweep table is empty at the beginning
        sweep_table = np.empty(BIA_SWEEP_TABLE_NR, dtype=np.float32)
        sweep_table[:] = np.nan

        # Loop over each row in the input data
        for current_row in data.to_dict(orient='records'):

            # Get packet name
            packet_name = current_row['palisade_id']

            # Make sure to start with an empty description
            description = {}
            if packet_name == 'TC_DPU_CLEAR_BIAS_SWEEP':
                # if valid clear table command is found, then
                # stores NULL values
                sweep_table[:] = np.nan
                description['sweep_step_cur'] = None
                description['sweep_eeprom'] = False
            elif packet_name == 'TC_DPU_LOAD_BIAS_SWEEP':
                # If valid load table command is executed,
                # then update the current sweep table values
                sweep_table = to_na(
                    current_row['data']['CP_DPU_BIA_SWEEP_FIRST_IDX'],
                    current_row['data']['CP_DPU_BIA_SWEEP_STEP_NR'],
                    current_row['data']['CP_DPU_BIA_SWEEP_STEP_CUR'],
                    sweep_table=sweep_table
                )
                # Store sweep table values in description
                description['sweep_step_na'] = list(sweep_table.astype(str))
                # Store EEPROM loading state
                description['sweep_eeprom'] = int(current_row['data'][
                    'CP_DPU_BIA_SWEEP_EEPROM']) == 1
            else:
                continue

            # Add a new entry for the event_log table
            new_entry = {
                'label': packet_name,
                'description': description,
                'start_time': current_row['utc_time'],
                'end_time': current_row['utc_time'],
                'is_predictive': True
            }
            output_data.append(new_entry)

        # Convert list of dictionaries into list of lists
        if len(output_data) > 0:
            # Store Bias sweep table values for bias sweep events (see in
            # self.biasweep_to_event)
            self.bia_sweep_tables = pd.DataFrame.from_records(
                output_data).sort_values(by=['start_time'], ignore_index=True)

        return self.bia_sweep_tables

    def biasweep_to_event(self, data):
        """
        Convert input data rows from pipeline.bia_sweep_log table
        into entries to be saved in pipeline.event_log table.

        :param data: input bia_sweep_log data rows (passed as a pandas.DataFrame object)
        :return: a pandas.DataFrame object contains sweeps to insert
        """
        # Initialize output
        output_data = []

        # Loops over the input data to get complete bias sweeps
        sweep_start_buffer = []
        ant_index_buffer = []
        for current_row in data.to_dict(orient='records'):
            if current_row['sweep_step'] in ['START_ANT1', 'START_ANT2', 'START_ANT3']:
                current_ant_index = current_row['sweep_step'][-1]
                if current_ant_index in ant_index_buffer:
                    current_sweep_index = ant_index_buffer.index(current_ant_index)
                    # If antenna index already found in buffer, this is not expected
                    logger.error(f"START_ANT{current_ant_index} already found in antenna index buffer. "
                                 f"Bias sweep on {sweep_start_buffer[current_sweep_index]} might be incomplete!")
                    # Remove already existing bia sweep start and antenna index values
                    # from the buffers
                    del sweep_start_buffer[current_sweep_index]
                    del ant_index_buffer[current_sweep_index]

                # Add the antenna index to the antenna index buffer list
                ant_index_buffer.append(current_ant_index)
                # add the sweep start time in sweep_start_buffer list
                sweep_start_buffer.append(current_row['utc_time'])
            elif current_row['sweep_step'] in ['END_ANT1', 'END_ANT2', 'END_ANT3']:
                # The END_ANT{ant_index} should match with ones of the previous START_ANT{ant_index} in ant_index_buffer
                # If not returns a warning message and skip sweep
                current_ant_index = current_row['sweep_step'][-1]
                if current_ant_index not in ant_index_buffer:
                    logger.warning(f'{current_row} is not associated with a previous '
                                   f'START_ANT{current_ant_index} event, skip sweep'
                                    )
                    continue
                else:
                    current_sweep_index = ant_index_buffer.index(current_ant_index)
                    current_sweep_start = sweep_start_buffer[current_sweep_index]
                    # Add a new entry for the event_log table
                    new_entry = {
                        'label': f'BIA_SWEEP_ANT{current_ant_index}',
                        'description': null(),
                        'start_time': current_sweep_start,
                        'end_time': current_row['utc_time'],
                        'is_predictive': current_row['utc_time_is_predictive']
                    }

                    # Retrieve values in nA of the Bias sweep table used for
                    # the current sweep
                    if self.bia_sweep_tables is not None:
                        current_table = self.bia_sweep_tables[
                            self.bia_sweep_tables.start_time <= current_sweep_start].sort_values(by=['start_time'],
                                                                                         ascending=False,
                                                                                         ignore_index=True)
                        new_entry['description'] = {
                            'sweep_step_na': current_table.description[0]['sweep_step_na']
                        }
                    else:
                        logger.warning(f'No valid Bias sweep table values found for current bias sweep!')

                    output_data.append(new_entry)

                # Remove bia sweep start and antenna index values
                # from the buffers
                del sweep_start_buffer[current_sweep_index]
                del ant_index_buffer[current_sweep_index]
            elif current_row['sweep_step'] in ['STEP_ANT1', 'STEP_ANT2', 'STEP_ANT3']:
                # Ignore detailed bias value steps
                pass
            else:
                pass

        # Convert list of dictionaries into list of lists
        if len(output_data) > 0:
            # Store bia sweep data as pandas.DataFrame object
            sweep_to_insert = pd.DataFrame.from_records(
                output_data).sort_values(by=['start_time'], ignore_index=True)
        else:
            sweep_to_insert = pd.DataFrame()

        return sweep_to_insert

    def tm_to_event(self, data):
        """
        Convert input data rows from pipeline.tm_log table
        into entries to be saved in pipeline.event_log table.

        :param data: input tm_log data rows as a pandas.DataFrame object
        :return: pandas.DataFrame object containing tm_log rows to insert
        """
        # Initialize output
        tm_to_insert = pd.DataFrame()

        tm_to_insert['label'] = data['palisade_id']
        tm_to_insert['description'] = data['data']
        tm_to_insert['start_time'] = data['utc_time']
        tm_to_insert['end_time'] = data['utc_time']
        tm_to_insert['is_predictive'] = data['utc_time_is_predictive']

        return tm_to_insert

    def tc_to_event(self, data):
        """
        Convert input data rows from pipeline.tc_log table
        into entries to be saved in pipeline.event_log table.

        :param data: input tc_log data rows passed a pandas.DataFrame object
        :return: pandas.DataFrame object containing tc_log rows to insert
        """
        # Initialize output
        tc_to_insert = pd.DataFrame()

        # Number of input rows
        n_data = data.shape[0]

        tc_to_insert['label'] = data['palisade_id']
        tc_to_insert['start_time'] = data['utc_time']
        tc_to_insert['end_time'] = data['utc_time']
        tc_to_insert['is_predictive'] = [True] * n_data

        # Set description field
        # By default store TC data in the description
        tc_to_insert['description'] = data['data']
        # For HFR FREQ LIST TC, provide a comprehensive list of frequency values
        # in the description
        where_hfr_list = data['palisade_id'].isin(HFR_FREQ_LIST_PACKETS)
        tc_to_insert.loc[where_hfr_list, ('description')] = data.loc[where_hfr_list].apply(
            lambda row: self.extract_hfr_freq_list(
                row['palisade_id'],
                row['data'],
            ), axis=1
        )
        # For Bias current TC, provide a comprehensive value of the bias
        # current in nA
        where_bia_cur = data['palisade_id'].isin(BIA_CURRENT_LOG_PACKETS)
        tc_to_insert.loc[where_bia_cur, ('description')] = data.loc[where_bia_cur].apply(
            lambda row: self.extract_bia_current(
                row['palisade_id'],
                row['data']
            ), axis=1)

        return tc_to_insert

    def lfrcalib_to_event(self, data):
        """
        Extract LFR calibration events from input TC data

        :param data: TC packet data rows (as a pandas.DataFrame Group)
        :return: pandas.DataFrame object containing the N LFR calibration events
        """

        # Initialize the default value of output
        calib_to_insert = pd.DataFrame()

        # Initialize loop variables
        calib_list = []
        current_calib = None

        # Loop over each row in the input data
        for current_row in data.to_dict(orient='records'):

            # Get packet name
            packet_name = current_row['palisade_id']

            if packet_name == 'TC_LFR_ENABLE_CALIBRATION' and not current_calib:
                # New LFR Calibration has started
                current_calib = current_row
            elif packet_name == 'TC_LFR_DISABLE_CALIBRATION' and current_calib:
                # Current LFR calibration has ended

                # Add a new entry for the event_log table
                new_entry = {
                    'label': 'LFR CALIBRATION',
                    'description': {'description': 'LFR internal calibration'},
                    'start_time': current_calib['utc_time'],
                    'end_time': current_row['utc_time'],
                    'is_predictive': True
                }
                calib_list.append(new_entry)

                # Reset current calibration event
                current_calib = None
            elif packet_name == 'TC_LFR_ENABLE_CALIBRATION' and current_calib:
                logger.warning(f'Incomplete LFR calibration for {current_calib}!')
                current_calib = None  # Force new LFR calibration event
            elif packet_name == 'TC_LFR_DISABLE_CALIBRATION' and not current_calib:
                logger.warning(f'Encounter TC_LFR_DISABLE_CALIBRATION, but not previous TC_LFR_ENABLE_CALIBRATION found!')

        # Convert list of dictionaries into pandas.DataFrame
        if len(calib_list) > 0:
            calib_to_insert = pd.DataFrame.from_records(calib_list)

        return calib_to_insert

    def extract_hfr_freq_list(self, packet_name, packet_data):
        """
        Extract the list of HFR frequencies from input TCs
        (for HFR LIST mode)

        :param packet_name: Name of the TC (PALISADE ID)
        :param packet_data: TC data (pandas.DataFrame)
        :return: hfr_freq_list dictionary with freq_list and freq_nr
        """
        hfr_freq_list = {}

        # Get mode and HFR band from packet_name
        # (e.g., TC_THR_LOAD_BURST_PAR_2)
        tc_fields = packet_name.split('_')
        # Get first letter of the current mode (N or B)
        mode = tc_fields[3][0]
        # Get HFR band
        band = int(tc_fields[-1][0]) - 1

        # Get number of frequencies to extract
        hfr_freq_list['freq_nr'] = int(packet_data[f'SY_THR_{mode}_SET_HLS{band}_NR_FREQ'])
        # Extract list of frequencies in kHz
        hfr_freq_list['freq_list'] = [
            str(compute_hfr_list_freq(current_freq)).strip()
            for current_freq in packet_data[f'SY_THR_{mode}_SET_HLS_FREQ_HF{band}']
        ]

        return hfr_freq_list

    def extract_bia_current(self, packet_name, packet_data):
        """
        Extract BIAS current value from input TC packet data

        :param packet_name: Name of the TC packet (palisade ID)
        :param packet_data: bias current TC data as a dictionary
        :return: Bias current value in nA set on current antenna
        """
        # Define CP_BIA_SET_BIAS{i} TC name (where i = [0, 1, 2])
        param_name = 'CP_BIA_SET_BIAS{0}'.format(packet_name[-1])
        # Get Bias current value in nA for the corresponding sensor
        # (requires IDB for transfer function)
        return {param_name: float(raw_to_na(packet_data[param_name],
                                            idb_source=self.idb_source,
                                            idb_version=self.idb_version))}

    def to_event_log(self, model, func,
                     filters=None):
        """
        Generic method to insert data into the pipeline.event_log table.

        :param model: table class of data to copy into event_log table
        :param func: method to use to map data between input and output table fields
        :param filters: list of filters for database query (time range filtering will be automatically appended)

        :return: inserted, failed and ignored entry counters + input data returned from database
        """
        # Get data from pipeline table
        data = self.query_data(model, filters=filters)

        n_data = data.shape[0]
        logger.info(f'{n_data} {model.__tablename__} data to insert into event_log '
                    f'between {self.start_time} and {self.end_time} for task EventToDb-{self.job_sid} ')
        if n_data == 0:
            return 0, 0, 0

        # Convert data into entries compatible with pipeline.event_log table
        data_to_insert = func(data)
        n_entry = data_to_insert.shape[0]

        # Insert data (or update if already exist)
        inserted_count = 0
        failed_count = 0
        ignored_count = 0
        if n_entry > 0:
            # Add insertion times
            data_to_insert['insert_time'] = [datetime.today()] * n_entry
            for i, current_entry in enumerate(self.df_to_dict(data_to_insert, EventLog)):
                if current_entry['label'] in [None, null()]:
                    logger.debug(f'{current_entry} is ignored for task EventToDb-{self.job_sid}')
                    ignored_count += 0
                    continue

                # Unique constraint is only on label, start_time and end_time
                update_fields_kwargs = {
                    key: val
                    for key, val in current_entry.items()
                    if key in ['start_time',
                               'end_time',
                               'label',
                               ]
                }
                if insert_in_db(self.session, EventLog, current_entry,
                                update_fields=current_entry,
                                update_fields_kwargs=update_fields_kwargs,
                                ) < 0:
                    logger.error(f'Inserting {current_entry} in pipeline.event_log table '
                                 f'has failed for task EventToDb-{self.job_sid}')
                    failed_count += 1
                else:
                    logger.debug(f'{current_entry} inserted into pipeline.event_log table for task EventToDb-{self.job_sid}')
                    inserted_count += 1

        return inserted_count, failed_count, ignored_count

    def df_to_dict(self, data, model,
                   remove=['id']):
        """
        Convert input pandas.Dataframe containing input data
        into dictionary.
        Keep only table columns as keywords in the output dictionary.

        :param data: pandas.Dataframe containing input data
        :param model: table class
        :param remove: list of fields to remove
        :return: dictionary
        """
        columns = get_columns(model, remove=remove)
        return data[columns].to_dict('records')

    def query_data(self, model, filters=[]):
        """
        Common method to query data from database and
        to return rows as pandas.Dataframe

        :param model: Table class
        :param filters: List of filters to query
        :return: pandas.Dataframe containing query results
        """

        # If not input filters, then filter at least by time range
        if filters is None:
            filters = gen_sql_filters(model,
                                      start_time=self.start_time,
                                      end_time=self.end_time)

        # If possible, sort returned rows by increasing UTC times
        try:
            order_by = model.utc_time
        except:
            order_by = None

        results = query_db(self.session, model,
                         filters=filters,
                         order_by=order_by,
                         tryouts=self.tryouts,
                         wait=self.wait,
                         limit=self.limit,
                         )
        return results



# -*- coding: utf-8 -*-
"""Modify file_log table


Revision ID: roc_dingo_0004_modify_file_log
Revises: roc_dingo_0003_add_cascade
Create Date: 2020-07-28 09:12:14.769013

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

from poppy.pop.alembic.helpers import create_table, drop_table, execute, user

# revision identifiers, used by Alembic.
revision = 'roc_dingo_0004_modify_file_log'
# down_revision = 'roc_dingo_0003_add_cascade'
down_revision = 'roc_dingo_0001_initial'
depends_on = None


def upgrade():
    # rename columns
    op.alter_column(
        'file_log', 'id_file_log',
        new_column_name='id', schema='pipeline')
    op.alter_column(
        'file_log', 'file_author',
        new_column_name='author', schema='pipeline')
    op.alter_column(
        'file_log', 'file_basename',
        new_column_name='basename', schema='pipeline')
    op.alter_column(
        'file_log', 'file_creation_date',
        new_column_name='creation_time', schema='pipeline')
    op.alter_column(
        'file_log', 'file_dataset_id',
        new_column_name='dataset_id', schema='pipeline')
    op.alter_column(
        'file_log', 'file_date',
        new_column_name='date', schema='pipeline')
    op.alter_column(
        'file_log', 'file_descr',
        new_column_name='descr', schema='pipeline')
    op.alter_column(
        'file_log', 'file_dir',
        new_column_name='dirname', schema='pipeline')
    op.alter_column(
        'file_log', 'file_endtime',
        new_column_name='end_time', schema='pipeline')
    op.alter_column(
        'file_log', 'file_insert_date',
        new_column_name='insert_time', schema='pipeline')
    op.alter_column(
        'file_log', 'file_sha',
        new_column_name='sha', schema='pipeline')
    op.alter_column(
        'file_log', 'file_size',
        new_column_name='size', schema='pipeline')
    op.alter_column(
        'file_log', 'file_starttime',
        new_column_name='start_time', schema='pipeline')
    op.alter_column(
        'file_log', 'file_state',
        new_column_name='state', schema='pipeline')
    op.alter_column(
        'file_log', 'file_status',
        new_column_name='status', schema='pipeline')
    op.alter_column(
        'file_log', 'file_url',
        new_column_name='url', schema='pipeline')
    op.alter_column(
        'file_log', 'file_version',
        new_column_name='version', schema='pipeline')

    op.drop_constraint(
        'file_log_file_basename_key', 'file_log',
        schema='pipeline', type_='unique')
    op.create_unique_constraint(
        'file_log_basename_key', 'file_log', ['basename'], schema='pipeline')

    # Very dirty hack because SA does not create
    # ENUM types when add_column, only with create_table
    # https://stackoverflow.com/questions/47206201/
    # how-to-use-enum-with-sqlalchemy-and-alembic
    # after deletting the table, the enum type is kept

    op.create_table(
        'dirty_hack',
        sa.Column('id', sa.BIGINT(), primary_key=True),
        sa.Column(
            'type',
            postgresql.ENUM(
                'CAL', 'HK', 'L0', 'L1', 'L1R', 'L2',
                'L3', 'LL01', 'TC', 'TM',
                name='file_type_list',
                )),
        schema='pipeline'
    )
    op.drop_table('dirty_hack', schema='pipeline')

    op.add_column('file_log', sa.Column(
        'type',
        postgresql.ENUM(
            'CAL', 'HK', 'L0', 'L1', 'L1R',
            'L2', 'L3', 'LL01', 'TC', 'TM',
            name='file_type_list'),
        autoincrement=False, nullable=False), schema='pipeline')
    op.add_column(
        'file_log', sa.Column('is_latest', sa.BOOLEAN(), nullable=False),
        schema='pipeline')
    op.add_column(
        'file_log', sa.Column('validity_end', postgresql.TIMESTAMP()),
        schema='pipeline')
    op.add_column(
        'file_log', sa.Column('validity_start', postgresql.TIMESTAMP()),
        schema='pipeline')

    op.drop_column('file_log', 'file_name', schema='pipeline')
    op.drop_column('file_log', 'is_test', schema='pipeline')
    op.drop_column('file_log', 'file_parents', schema='pipeline')
    op.drop_column('file_log', 'file_id', schema='pipeline')
    op.drop_column('file_log', 'file_level', schema='pipeline')

    # Start and stop time can be null
    op.alter_column(
        'file_log', 'start_time',
        nullable=True, schema='pipeline')
    op.alter_column(
        'file_log', 'end_time',
        nullable=True, schema='pipeline')

    # create parents table
    create_table(
        'parents_file_log',
        sa.Column('id', sa.BIGINT(), nullable=False),
        sa.Column('id_child', sa.BIGINT(), nullable=False),
        sa.Column('id_parent', sa.BIGINT(), nullable=False),
        sa.ForeignKeyConstraint(
            ['id_child'], ['pipeline.file_log.id'], ondelete='CASCADE'),
        sa.ForeignKeyConstraint(
            ['id_parent'], ['pipeline.file_log.id'], ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('id_child', 'id_parent'),
        schema='pipeline'
    )
    # ### end Alembic commands ###

    # fix table permissions
    grant_user_seq = (
        """GRANT USAGE, SELECT
        ON ALL SEQUENCES IN SCHEMA pipeline TO {0}""".format(user)
    )
    execute(grant_user_seq)


def downgrade():
    # first drop parents_file_log table
    drop_table('parents_file_log', schema='pipeline')

    # Start and stop time cannot be null
    op.alter_column(
        'file_log', 'start_time',
        nullable=False, schema='pipeline')
    op.alter_column(
        'file_log', 'end_time',
        nullable=False, schema='pipeline')

    # rename columns

    op.alter_column(
        'file_log', 'id',
        new_column_name='id_file_log', schema='pipeline')
    op.alter_column(
        'file_log', 'author',
        new_column_name='file_author', schema='pipeline')
    op.alter_column(
        'file_log', 'basename',
        new_column_name='file_basename', schema='pipeline')
    op.alter_column(
        'file_log', 'creation_time',
        new_column_name='file_creation_date', schema='pipeline')
    op.alter_column(
        'file_log', 'dataset_id',
        new_column_name='file_dataset_id', schema='pipeline')
    op.alter_column(
        'file_log', 'date',
        new_column_name='file_date', schema='pipeline')
    op.alter_column(
        'file_log', 'descr',
        new_column_name='file_descr', schema='pipeline')
    op.alter_column(
        'file_log', 'dirname',
        new_column_name='file_dir', schema='pipeline')
    op.alter_column(
        'file_log', 'end_time',
        new_column_name='file_endtime', schema='pipeline')
    op.alter_column(
        'file_log', 'insert_time',
        new_column_name='file_insert_date', schema='pipeline')
    op.alter_column(
        'file_log', 'sha',
        new_column_name='file_sha', schema='pipeline')
    op.alter_column(
        'file_log', 'size',
        new_column_name='file_size', schema='pipeline')
    op.alter_column(
        'file_log', 'start_time',
        new_column_name='file_starttime', schema='pipeline')
    op.alter_column(
        'file_log', 'state',
        new_column_name='file_state', schema='pipeline')
    op.alter_column(
        'file_log', 'status',
        new_column_name='file_status', schema='pipeline')
    op.alter_column(
        'file_log', 'url',
        new_column_name='file_url', schema='pipeline')
    op.alter_column(
        'file_log', 'version',
        new_column_name='file_version', schema='pipeline')

    op.drop_constraint(
        'file_log_basename_key', 'file_log',
        schema='pipeline', type_='unique')
    op.create_unique_constraint(
        'file_log_file_basename_key', 'file_log',
        ['file_basename'], schema='pipeline')

    op.drop_column('file_log', 'type', schema='pipeline')
    op.drop_column('file_log', 'is_latest', schema='pipeline')
    op.drop_column('file_log', 'validity_end', schema='pipeline')
    op.drop_column('file_log', 'validity_start', schema='pipeline')

    execute('DROP TYPE IF EXISTS file_type_list')

    op.add_column(
        'file_log',
        sa.Column('file_id', sa.String(length=48), nullable=False),
        schema='pipeline')
    op.add_column(
        'file_log',
        sa.Column('file_level', sa.String(length=8), nullable=False),
        schema='pipeline')
    op.add_column(
        'file_log',
        sa.Column(
            'file_parents', sa.VARCHAR(length=512),
            autoincrement=False, nullable=True),
        schema='pipeline')
    op.add_column(
        'file_log',
        sa.Column(
            'is_test', sa.BOOLEAN(),
            autoincrement=False, nullable=False),
        schema='pipeline')
    op.add_column(
        'file_log', sa.Column(
            'file_name', sa.VARCHAR(length=512),
            autoincrement=False, nullable=True),
        schema='pipeline')

    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
"""empty message

Revision ID: roc_dingo_0025_add_views_for_anomalies
Revises: roc_dingo_0024_add_invalidpacketlog
Create Date: 2022-07-28 09:35:17.202789

"""
from alembic import op
import sqlalchemy as sa
from pathlib import Path

# revision identifiers, used by Alembic.
revision = 'roc_dingo_0025_add_views_for_anomalies'
down_revision = 'roc_dingo_0024_add_invalidpacketlog'
branch_labels = None
depends_on = None


def upgrade():
    # ensure commands are executed now
    # op.execute('COMMIT')

    op.create_table('anomalies',
                    sa.Column('id', sa.BIGINT(), nullable=False),
                    sa.Column('label', sa.String(), nullable=False),
                    sa.Column('tracked', sa.BOOLEAN(), nullable=False),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('label'),
                    schema='pipeline')

    # set default value to TRUE
    op.alter_column('anomalies', 'tracked',
                    server_default=sa.schema.DefaultClause('1'),
                    schema='pipeline')

    # insert data into anomalies
    filename = 'SQL/roc_dingo_0025_insert_into_anomalies.sql'
    sql_insert_file = Path(__file__).parent / filename
    with open(sql_insert_file, 'r') as sql_file:
        sql_cmd = sql_file.read()
        op.execute(sql_cmd)

    # create views
    filename = 'SQL/roc_dingo_0025_create_view_anomalies_per_day.sql'
    sql_file = Path(__file__).parent / filename
    with open(sql_file, 'r') as sql_content:
        sql_cmd = sql_content.read()
        op.execute(sql_cmd)

    filename = 'SQL/roc_dingo_0025_create_view_total_events.sql'
    sql_file = Path(__file__).parent / filename
    with open(sql_file, 'r') as sql_content:
        sql_cmd = sql_content.read()
        op.execute(sql_cmd)


def downgrade():
    # op.execute('COMMIT')
    op.execute('DROP VIEW IF EXISTS pipeline.total_events')
    op.execute('DROP VIEW IF EXISTS pipeline.anomalies_per_day')
    op.drop_table('anomalies', schema='pipeline')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Search for items in file_log to be updated
"""

import os

from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import load_only

from datetime import datetime

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task
from poppy.core.target import BaseTarget

from roc.dingo.constants import \
    PIPELINE_DATABASE, \
    ROC_DATA_ROOT
from roc.dingo.models.file import FileLog

__all__ = ['SearchForUpdate']


class SearchForUpdate(Task):
    """
    Parse ROC file tree and synchronize with the the ROC database
    """
    plugin_name = 'roc.dingo'
    name = 'search_for_update'
    files_to_check = []
    files_to_insert = []
    files_to_update = []

    def add_targets(self):
        logger.debug('SearchForUpdate() : add_targets')

        self.add_output(target_class=BaseTarget, many=True,
                        identifier='roc_data_files_to_update')
        self.add_output(target_class=BaseTarget, many=True,
                        identifier='roc_data_files_to_insert')

    def setup_inputs(self):
        """
        Setup task inputs.

        :param task:
        :return:
        """

        logger.debug('SearchForUpdate() : setup_inputs')

        # get the root file tree
        self.root = self.pipeline.get('root', default=ROC_DATA_ROOT, args=True)
        # ensure that there is / at the end
        self.root = os.path.join(self.root, '')

    def run(self):

        logger.debug('SearchForUpdate() : run')

        # Get the database connection if needed
        if not hasattr(self, 'session'):
            self.session = Connector.manager[PIPELINE_DATABASE].session

        # Initialize inputs
        self.setup_inputs()

        # Initialize outputs
        self.files_to_update = {}

        # Define filters to check file_log entry to update
        update_filter = FileLog.to_update == True  # noqa: E712

        # Define relevant fields to retreive
        fields = ['basename', 'dirname', 'creation_time', 'size']

        try:
            query = self.session.query(FileLog).\
                options(load_only(*fields)).\
                filter(update_filter)
            logger.debug(
                str(query.statement.compile(
                    dialect=postgresql.dialect(),
                    compile_kwargs={'literal_binds': True})))
            results = query.all()
        except Exception:  # pragma: no cover
            logger.error('Query failed :')
            logger.error(
                str(query.statement.compile(
                    dialect=postgresql.dialect(),
                    compile_kwargs={'literal_binds': True})))
        else:
            for item in results:
                full_name = os.path.join(
                    self.root, item.dirname, item.basename)

                # test if the file is still here
                if os.path.exists(full_name):
                    # do not rely on database creation_time and size
                    size = os.stat(full_name).st_size,
                    creation_time = datetime.utcfromtimestamp(
                        int(os.path.getmtime(full_name)))
                else:
                    # it was removed, keep the values in db
                    logger.info(f'File {item.basename} is not anymore '
                                'in the file system -> mark as removed')
                    item.is_removed = True
                    item.state = 'OK'
                    item.error_log = None
                    item.to_update = False
                    self.session.commit()
                    # do not continue with this file
                    continue

                self.files_to_update[item.basename] = {
                    'filepath': os.path.join(
                        self.root,
                        item.dirname,
                        item.basename),
                    'creation_time': creation_time,
                    'size': size,
                }

        # Returns outputs
        self.outputs['roc_data_files_to_update'].data = self.files_to_update
        logger.info(
            '{} files to update'.format(str(len(self.files_to_update))))
        if len(self.files_to_update) == 0:
            self.outputs['roc_data_files_to_update'].is_empty = True

        self.outputs['roc_data_files_to_insert'].data = {}
        self.outputs['roc_data_files_to_insert'].is_empty = True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Reference documentation
# ROC-GEN-SYS-NTT-00038-LES_Iss01_Rev02(Mission_Database_Description_Document)

"""
Database model for rpw file processing history tables.
"""

from sqlalchemy import ForeignKey, String, Table, Column
from sqlalchemy.orm import relationship
from sqlalchemy.dialects.postgresql import \
    BIGINT, INTEGER, TIMESTAMP, ENUM, BOOLEAN

from poppy.core.db.base import Base
from poppy.core.db.non_null_column import NonNullColumn

__all__ = [
    'FileLog',
]


# File lists
FILE_STATE_LIST = ['OK', 'WARNING', 'ERROR']
FILE_STATUS_LIST = ['Pended', 'Terminated', 'InProgress', 'Locked', 'Failed']
FILE_LEVEL_LIST = [
    'CAL', 'TC', 'TM', 'SK', 'L0', 'HK', 'L1',
    'BIA', 'L1R', 'L2', 'L3', 'LL01']

# file_action_list = ['Process', 'Reprocess', 'Lock', 'Unlock', 'Delete']

file_state_enum = ENUM(
    *FILE_STATE_LIST, name='file_state_list')
file_status_enum = ENUM(
    *FILE_STATUS_LIST, name='file_status_list')
file_level_enum = ENUM(
    *FILE_LEVEL_LIST, name='file_level_list')

ParentsFileLog = Table(
    'parents_file_log',
    Base.metadata,
    Column('id_parent', INTEGER, ForeignKey('pipeline.file_log.id')),
    Column('id_child', INTEGER, ForeignKey('pipeline.file_log.id')),
    schema='pipeline'
)


class FileLog(Base):
    """
    Class representation of the table for file_log table in the ROC database.
    """

    id = NonNullColumn(
        BIGINT(),
        primary_key=True)
    sha = NonNullColumn(
        String(64),
        nullable=True,
        descr='SHA256 of the file')
    basename = NonNullColumn(
        String(512),
        descr='Basename of the file',
        unique=True)
    product = NonNullColumn(
        String(512),
        descr='Basename of the file without extension and version number')
    version = NonNullColumn(
        String(16),
        descr='Data version of the file')
    state = NonNullColumn(
        file_state_enum,
        descr=f'State of the file. Possible values are: {FILE_STATE_LIST}')
    status = NonNullColumn(
        file_status_enum,
        descr=f'Status of the file. Possible values are: {FILE_STATUS_LIST}')
    level = NonNullColumn(
        file_level_enum,
        descr=f'Level of the file. Possible values are: {FILE_LEVEL_LIST}')
    creation_time = NonNullColumn(
        TIMESTAMP(),
        nullable=True,
        descr='Local date and time of the file creation',
        comment='')
    insert_time = NonNullColumn(
        TIMESTAMP(),
        descr='Local date and time of the file insertion in the database',
        nullable=True)
    descr = NonNullColumn(
        String(512),
        nullable=True,
        descr='Description of the file')
    author = NonNullColumn(
        String(512),
        nullable=True,
        descr='Author of the file')
    dirname = NonNullColumn(
        String(512),
        nullable=True,
        descr='Relative path of the file directory (from data root base)')
    url = NonNullColumn(
        String(512),
        nullable=True,
        descr='URL of the file in the HTTPS server')
    size = NonNullColumn(
        BIGINT(),
        nullable=True,
        descr='Size of the file in kilobytes')
    start_time = NonNullColumn(
        TIMESTAMP(),
        nullable=True,
        descr='Start time of file')
    end_time = NonNullColumn(
        TIMESTAMP(),
        nullable=True,
        descr='End time of file')
    validity_start = NonNullColumn(
        TIMESTAMP(),
        nullable=True,
        descr='Start time of validity range (for CAL files)')
    validity_end = NonNullColumn(
        TIMESTAMP(),
        nullable=True,
        descr='End time of validity range (for CAL files)')
    dataset_id = NonNullColumn(
        String(512),
        nullable=True,
        descr='Dataset ID in the ROC database')
    is_archived = NonNullColumn(
        BOOLEAN,
        default=False,
        descr='True if file has been archived at ESAC')
    is_delivered = NonNullColumn(
        BOOLEAN,
        default=False,
        descr='True if file has been delivered ESAC')
    public_filename = NonNullColumn(
        String(512),
        descr='Public filename of the file delivered to ESAC',
        nullable=True,
        unique=False)
    public_dirname = NonNullColumn(
        String(512),
        descr='Public dirname of the file delivered to ESAC',
        nullable=True,
        unique=False)
    is_latest = NonNullColumn(
        BOOLEAN,
        default=False,
        descr='True if the version is the latest one')
    is_removed = NonNullColumn(
        BOOLEAN,
        default=False,
        descr='True if the file has been removed from the file directory')
    error_log = NonNullColumn(
        String(),
        nullable=True,
        descr='Log message when an error occured during insertion')
    to_update = NonNullColumn(
        BOOLEAN,
        default=False,
        descr='True if file has to be updated by DINGO')

    parents = relationship(
        'FileLog',
        secondary=ParentsFileLog,
        primaryjoin=ParentsFileLog.c.id_child == id,
        secondaryjoin=ParentsFileLog.c.id_parent == id,
        backref='children',
        cascade='all, delete',
    )

    __tablename__ = 'file_log'
    __table_args__ = (
        {
            'schema': 'pipeline',
        }
    )

    def as_dict(self):
        # init the dictionnary
        log_file_dict = {
            c.name: getattr(self, c.name)
            for c in self.__table__.columns
        }

        # delete some useless columns
        log_file_dict.pop('id')
        log_file_dict.pop('insert_time')

        # datetime formatting
        fields = ['creation_time', 'start_time', 'end_time',
                  'validity_start', 'validity_end']
        for f in fields:
            time = log_file_dict[f]
            if time is not None:
                log_file_dict[f] = time.isoformat()

        log_file_dict['parents'] = []
        for p in self.parents:
            log_file_dict['parents'].append(p.basename)

        log_file_dict['parents'].sort()

        return log_file_dict



"""empty message

Revision ID: roc_dingo_0023_add_hfrtimelog
Revises: roc_dingo_0022_add_to_update_flag_to_filelog
Create Date: 2022-02-08 17:22:38.827046

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

from poppy.pop.alembic.helpers import \
    create_table, execute, user, drop_table

# revision identifiers, used by Alembic.
revision = 'roc_dingo_0023_add_hfrtimelog'
down_revision = 'roc_dingo_0022_add_to_update_flag_to_filelog'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    create_table('hfr_time_log',
    sa.Column('id', sa.BIGINT(), nullable=False),
    sa.Column('acq_time', postgresql.TIMESTAMP(), nullable=False),
    sa.Column('coarse_time', sa.BIGINT(), nullable=False),
    sa.Column('fine_time', sa.BIGINT(), nullable=False),
    sa.Column('mode', sa.SMALLINT(), nullable=False),
    sa.Column('delta_time1', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
    sa.Column('delta_time2', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('coarse_time', 'fine_time', 'mode'),
    schema='pipeline'
    )
    # ### end Alembic commands ###

    # added by hand
    # fix table permissions
    grant_user_seq = (
        """GRANT USAGE, SELECT
        ON ALL SEQUENCES IN SCHEMA pipeline TO {0}""".format(user)
    )
    execute(grant_user_seq)

def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    drop_table('hfr_time_log', schema='pipeline')
    # ### end Alembic commands ###



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Tests module for the roc.dingo plugin.
"""
import pytest
import tempfile
import shutil
import os.path as osp
from os import listdir
from roc.dingo.models.file import FileLog
from roc.dingo.tests.test_log_file import \
    DingoLogFileCommandTestCase, DingoLogFileTest


# Tests on roc.dingo.tasks.log_file methods
class TestDingoStoreDelivered(DingoLogFileCommandTestCase):

    def setup_method(self, method):
        super().setup_method(method)

        # empty the database
        self.session.query(FileLog).delete()
        self.session.flush()

    def teardown_method(self, method):
        super().teardown_method(method)

        # close database
        self.session.close()

    @pytest.mark.parametrize(
        'root, public_root, delivered_root, remove_files, '
        'nb_files_after_logfile_to_rocdb, '
        'nb_files_after_store_delivered, '
        'nb_public_files, nb_delivered_files', [
            ('DATA-1', 'PUBLIC-1', 'DELIVERED-1', True, 61, 106, 76, 72),
            ('DATA-1', 'PUBLIC-1', 'DELIVERED-1', False, 61, 106, 76, 72),
        ])
    def test_store_delivered(self,
                             root,
                             public_root,
                             delivered_root,
                             remove_files,
                             nb_files_after_logfile_to_rocdb,
                             nb_files_after_store_delivered,
                             nb_public_files,
                             nb_delivered_files):
        """
        Insert dataset with logfile_to_rocdb and store delivered file after
        """
        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)
        public_root = osp.join(data_test_path, public_root)
        delivered_root = osp.join(data_test_path, delivered_root)

        # --- initialize the Db with files from DATA-1 ---
        main_command = ['pop',
                        'dingo',
                        'logfile_to_rocdb',
                        '--root',
                        root
                        ]

        # --- run the command ---
        self.run_command(main_command)

        results = self.session.query(FileLog).all()
        assert len(results) == nb_files_after_logfile_to_rocdb

        # No files should be tagged is_removed
        results = self.session.query(FileLog).\
            filter(FileLog.is_removed == True).all()  # noqa: E712
        assert len(results) == 0

        # As the store_delivered command will delete file,
        # copy the folder to a tmp one

        tmp_delivered_root = tempfile.mkdtemp()
        for f in listdir(delivered_root):
            shutil.copy(osp.join(delivered_root, f), tmp_delivered_root)

        # Check the copy
        results = listdir(tmp_delivered_root)
        assert len(results) == nb_delivered_files

        # --- insert delivered data ---
        main_command = ['pop',
                        'dingo',
                        'store_delivered',
                        '--public',
                        public_root,
                        '--delivered',
                        tmp_delivered_root
                        ]

        if remove_files:
            main_command += ['--remove_files']

        # --- run the command ---
        self.run_command(main_command)

        # --- make assertions ---

        # Nb files after store_delivered
        results = self.session.query(FileLog).all()
        assert len(results) == nb_files_after_store_delivered

        # All the files added should be tagged is_removed
        results = self.session.query(FileLog).\
            filter(FileLog.is_removed == True).all()  # noqa: E712
        assert len(results) == \
            nb_files_after_store_delivered - nb_files_after_logfile_to_rocdb

        # Check the nb of files with a public file mentionned
        results = self.session.query(FileLog).\
            filter(FileLog.public_filename.isnot(None)).all()
        assert len(results) == nb_public_files

        # Check the nb of delivered files
        results = self.session.query(FileLog).\
            filter(FileLog.is_delivered == True).all()  # noqa: E712
        assert len(results) == nb_delivered_files

        # Check all the delivered files have been removed
        # if remove_files has been set
        results = listdir(tmp_delivered_root)
        if remove_files:
            assert len(results) == 0
        else:
            assert len(results) == nb_delivered_files

        # remove temporary directory
        shutil.rmtree(tmp_delivered_root)

        self.session.close()



# -*- coding: utf-8 -*-
"""empty message

Revision ID: roc_dingo_0001_initial
Revises: poppy_pop_0001_initial
Create Date: 2020-02-10 13:59:42.454002

"""
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql
from alembic import op

from poppy.pop.alembic.helpers import \
    create_table, create_schema, execute, user, drop_table

# revision identifiers, used by Alembic.
revision = 'roc_dingo_0001_initial'
down_revision = None
branch_labels = None
depends_on = 'poppy_pop_0002_version_num'


def upgrade():

    create_schema('pipeline')  # added by hand

    ## delete the types if they exist
    execute('DROP TYPE IF EXISTS file_state_type')
    execute('DROP TYPE IF EXISTS file_status_type')

    # ### commands auto generated by Alembic - please adjust! ###
    create_table(
        'file_log',
        sa.Column('id_file_log', sa.BIGINT(), nullable=False),
        sa.Column('file_sha', sa.String(length=64), nullable=True),
        sa.Column('file_id', sa.String(length=48), nullable=False),
        sa.Column('file_basename', sa.String(length=512), nullable=False),
        sa.Column('file_version', sa.String(length=16), nullable=False),
        sa.Column(
            'file_state',
            postgresql.ENUM('OK', 'WARNING', 'ERROR',
                            name='file_state_type'),
            nullable=False),
        sa.Column(
            'file_status',
            postgresql.ENUM(
                'Pended', 'Terminated', 'InProgress', 'Locked', 'Failed',
                name='file_status_type'),
            nullable=False),
        sa.Column('file_creation_date', postgresql.TIMESTAMP(), nullable=True),
        sa.Column('file_insert_date', postgresql.TIMESTAMP(), nullable=True),
        sa.Column('file_descr', sa.String(length=512), nullable=True),
        sa.Column('file_name', sa.String(length=512), nullable=True),
        sa.Column('file_author', sa.String(length=512), nullable=True),
        sa.Column('file_dir', sa.String(length=512), nullable=True),
        sa.Column('file_url', sa.String(length=512), nullable=True),
        sa.Column('file_size', sa.BIGINT(), nullable=True),
        sa.Column('file_level', sa.String(length=8), nullable=True),
        sa.Column('file_date', sa.DATE(), nullable=True),
        sa.Column('file_starttime', postgresql.TIMESTAMP(), nullable=False),
        sa.Column('file_endtime', postgresql.TIMESTAMP(), nullable=False),
        sa.Column('file_dataset_id', sa.String(length=512), nullable=True),
        sa.Column('is_archived', sa.BOOLEAN(), nullable=False),
        sa.Column('is_test', sa.BOOLEAN(), nullable=False),
        sa.Column('file_parents', sa.String(length=512), nullable=True),
        sa.PrimaryKeyConstraint('id_file_log'),
        sa.UniqueConstraint('file_basename'),
        schema='pipeline'
    )
    create_table(
        'packet_log',
        sa.Column('id_packet_log', sa.BIGINT(), nullable=False),
        sa.Column('length', sa.SMALLINT(), nullable=False),
        sa.Column('type', sa.String(length=8), nullable=False),
        sa.Column('category', sa.String(length=512), nullable=False),
        sa.Column('apid', sa.SMALLINT(), nullable=True),
        sa.Column('sync_flag', sa.BOOLEAN(), nullable=True),
        sa.Column('utc_time', postgresql.TIMESTAMP(), nullable=False),
        sa.Column('srdb_id', sa.String(length=16), nullable=False),
        sa.Column('palisade_id', sa.String(length=256), nullable=False),
        sa.Column('binary', sa.String(), nullable=False),
        sa.Column('sha', sa.String(), nullable=True),
        sa.Column('idb_version', sa.String(length=128), nullable=True),
        sa.Column('idb_source', sa.String(length=128), nullable=True),
        sa.Column('creation_time', sa.String(length=1024), nullable=True),
        sa.Column('ack_exe_state', sa.String(length=16), nullable=True),
        sa.Column('ack_acc_state', sa.String(length=16), nullable=True),
        sa.Column('sequence_name', sa.String(length=16), nullable=True),
        sa.Column('unique_id', sa.String(length=256), nullable=True),
        sa.Column('insertion_time', postgresql.TIMESTAMP(), nullable=False),
        sa.PrimaryKeyConstraint('id_packet_log'),
        sa.UniqueConstraint('utc_time', 'srdb_id'),
        schema='pipeline'
    )
    create_table(
        'invalid_packet_log',
        sa.Column('id_invalid_packet_log', sa.BIGINT(), nullable=False),
        sa.Column('binary', sa.String(), nullable=False),
        sa.Column('insertion_time', postgresql.TIMESTAMP(), nullable=False),
        sa.Column('apid', sa.SMALLINT(), nullable=True),
        sa.Column('srdb_id', sa.String(length=16), nullable=True),
        sa.Column('sha', sa.String(), nullable=True),
        sa.Column('utc_time', postgresql.TIMESTAMP(), nullable=True),
        sa.Column('idb_version', sa.String(length=128), nullable=True),
        sa.Column('idb_source', sa.String(length=128), nullable=True),
        sa.Column('comment', sa.String(), nullable=True),
        sa.PrimaryKeyConstraint('id_invalid_packet_log'),
        sa.UniqueConstraint('binary'),
        schema='pipeline'
    )
    # ### end Alembic commands ###

    # added by hand
    # fix table permissions
    grant_user_seq = (
        """GRANT USAGE, SELECT
        ON ALL SEQUENCES IN SCHEMA pipeline TO {0}""".format(user)
    )
    execute(grant_user_seq)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    drop_table('packet_log', schema='pipeline')
    drop_table('file_log', schema='pipeline')
    drop_table('invalid_packet_log', schema='pipeline')
    # ### end Alembic commands ###

    ## delete the types if they exist
    execute('DROP TYPE IF EXISTS file_state_type')
    execute('DROP TYPE IF EXISTS file_status_type')

    # Delete pipeline schema
    execute('DROP SCHEMA pipeline')




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Tests module for the roc.dingo plugin.
"""
import pytest
import os.path as osp
from argparse import Namespace

from poppy.core.pipeline import Pipeline
from poppy.core.logger import logger

from roc.dingo.tests.test_log_file import \
    DingoLogFileTaskTestCase, DingoLogFileTest


# Tests on roc.dingo.tasks methods
class TestDingoLogFileParseFileTree(DingoLogFileTaskTestCase):

    @pytest.mark.parametrize('root, dir_file, expected_data_file', [
        ('DATA-1/', 'roc_dir_list_all.dat', 'roc_data_files_all.dat'),
    ])
    def test_parse_file_tree(self, root, dir_file, expected_data_file):
        """
        Test the DINGO ParseFileTree task with standard parameters

        :return:
        """

        # --- initialize the task ---
        from roc.dingo.tasks import ParseFileTree

        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)
        args = Namespace(root=root, dry_run=False)

        # --- initialise task & pipeline --
        self.task = ParseFileTree()
        self.task.pipeline = Pipeline(args)

        # --- run the task ---
        self.task.instantiate_targets()
        self.run_task()

        # --- get results ---
        roc_dir_list = self.task.outputs['roc_dir_list'].data
        roc_data_files = self.task.outputs['roc_data_files'].data

        # --- make assertions ---
        # --- DIR LIST ---

        # open the dir_list file with data for test
        roc_dir_list_all = DingoLogFileTest.read_data(dir_file)

        # check content
        assert roc_dir_list == roc_dir_list_all, \
            'dir list is not the one expected'

        # --- FILE LIST ---
        # read expected data
        expected_data = DingoLogFileTest.open_test_data(
            expected_data_file, root)

        # compare results with expected data
        DingoLogFileTest.compare_data_to_expected(
            expected_data, roc_data_files)

    @pytest.mark.parametrize('root, date, dir_file, expected_data_file', [
        ('DATA-1/',
         '2020/08/19',
         'roc_dir_list_2020_08_19.dat',
         'roc_data_files_2020_08_19.dat'),
    ])
    def test_parse_file_tree_with_day_date(
            self, root, date, dir_file, expected_data_file):
        """
        Test the DINGO ParseFileTree task with a day date parameter
        All files set in a month or mission basis should be excluded
        :return:
        """

        # --- initialize the task ---
        from roc.dingo.tasks import ParseFileTree

        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)
        args = Namespace(root=root, date=date, dry_run=False)

        # --- initialise task & pipeline --
        self.task = ParseFileTree()
        self.task.pipeline = Pipeline(args)

        # --- run the task ---
        self.task.instantiate_targets()
        self.run_task()

        # --- get results ---
        roc_dir_list = self.task.outputs['roc_dir_list'].data
        roc_data_files = self.task.outputs['roc_data_files'].data

        logger.debug('--- results ---')
        logger.debug(len(roc_data_files))
        for key, item in roc_data_files.items():
            logger.debug(item['filepath'].replace(root, ''))

        # --- make assertions ---
        # --- DIR LIST ---

        # open the dir_list file with data for test
        roc_dir_list_all = DingoLogFileTest.read_data(dir_file)

        # check content
        assert roc_dir_list == roc_dir_list_all, \
            'dir list is not the one expected'

        # --- FILE LIST ---
        # read expected data
        expected_data = DingoLogFileTest.open_test_data(
            expected_data_file, root)

        # compare resulsts with expected data
        DingoLogFileTest.compare_data_to_expected(
            expected_data, roc_data_files)

    @pytest.mark.parametrize('root, date, dir_file', [
        ('DATA-1/',
         '2020/08/18',
         'roc_dir_list_2020_08_18.dat'),
    ])
    def test_parse_file_tree_with_another_day_date(
            self, root, date, dir_file):
        """
        Test the DINGO ParseFileTree task with a another day date parameter
        File list should be empty
        :return:
        """

        # --- initialize the task ---
        from roc.dingo.tasks import ParseFileTree

        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)
        args = Namespace(root=root, date=date, dry_run=False)

        # --- initialise task & pipeline --
        self.task = ParseFileTree()
        self.task.pipeline = Pipeline(args)

        # --- run the task ---
        self.task.instantiate_targets()
        self.run_task()

        # --- get results ---
        roc_dir_list = self.task.outputs['roc_dir_list'].data
        roc_data_files = self.task.outputs['roc_data_files'].data

        logger.debug('--- results ---')
        logger.debug(len(roc_data_files))
        for key, item in roc_data_files.items():
            logger.debug(item['filepath'].replace(root, ''))

        # --- make assertions ---
        # --- DIR LIST ---

        # open the dir_list file with data for test
        roc_dir_list_all = DingoLogFileTest.read_data(dir_file)

        # check content
        assert roc_dir_list == roc_dir_list_all, \
            'dir list is not the one expected'

        # --- FILE LIST ---
        # read expected data
        expected_data = {}

        # compare resulsts with expected data
        DingoLogFileTest.compare_data_to_expected(
            expected_data, roc_data_files)

    @pytest.mark.parametrize('root, date, dir_file, expected_data_file', [
        ('DATA-1/',
         '2020/08',
         'roc_dir_list_2020_08.dat',
         'roc_data_files_2020_08.dat'),
    ])
    def test_parse_file_tree_with_month_date(
            self, root, date, dir_file, expected_data_file):
        """
        Test the DINGO ParseFileTree task with a month date parameter
        All files set in the mission basis should be excluded
        :return:
        """

        # --- initialize the task ---
        from roc.dingo.tasks import ParseFileTree

        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)
        args = Namespace(root=root, date=date, dry_run=False)

        # --- initialise task & pipeline --
        self.task = ParseFileTree()
        self.task.pipeline = Pipeline(args)

        # --- run the task ---
        self.task.instantiate_targets()
        self.run_task()

        # --- get results ---
        roc_dir_list = self.task.outputs['roc_dir_list'].data
        roc_data_files = self.task.outputs['roc_data_files'].data

        logger.debug('--- results ---')
        logger.debug(len(roc_data_files))
        for key, item in roc_data_files.items():
            logger.debug(item['filepath'].replace(root, ''))

        # --- make assertions ---
        # --- DIR LIST ---

        # open the dir_list file with data for test
        roc_dir_list_all = DingoLogFileTest.read_data(dir_file)

        # check content
        assert roc_dir_list == roc_dir_list_all, \
            'dir list is not the one expected'

        # --- FILE LIST ---
        # read expected data
        expected_data = DingoLogFileTest.open_test_data(
            expected_data_file, root)

        # compare resulsts with expected data
        DingoLogFileTest.compare_data_to_expected(
            expected_data, roc_data_files)

    @pytest.mark.parametrize('root, expected_data_file', [
        ('DATA-1/', 'roc_data_files_all.dat')
    ])
    def test_parse_file_tree_nf(self, root, expected_data_file):
        """
        Test the DINGO ParseFileTree task with the option -nf
        files from "former_versions" should be excluded

        :return:
        """

        # --- initialize the task ---
        from roc.dingo.tasks import ParseFileTree

        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)
        args = Namespace(root=root, no_former_versions=True, dry_run=False)

        # --- initialise task & pipeline --
        self.task = ParseFileTree()
        self.task.pipeline = Pipeline(args)

        # --- run the task ---
        self.task.instantiate_targets()
        self.run_task()

        # --- get results ---
        roc_data_files = self.task.outputs['roc_data_files'].data

        # --- make assertions ---
        # useless to test directory here
        # --- FILE LIST ---
        expected_data = DingoLogFileTest.open_test_data(
            expected_data_file, root)
        # remove "former_versions" items
        for key, item in expected_data.copy().items():
            if 'former_versions' in item['filepath']:
                expected_data.pop(key)

        # compare resulsts with expected data
        DingoLogFileTest.compare_data_to_expected(
            expected_data, roc_data_files)

    @pytest.mark.parametrize('root, expected_data_file', [
        ('SK-1/', 'roc_sk_files_1.dat'),
    ])
    def test_parse_file_tree_sk(self, root, expected_data_file):
        """
        Test the DINGO ParseFileTree task for spice kernels

        :return:
        """

        # --- initialize the task ---
        from roc.dingo.tasks import ParseFileTree

        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)
        args = Namespace(root=root, level='SK', dry_run=False)

        # --- initialise task & pipeline --
        self.task = ParseFileTree()
        self.task.pipeline = Pipeline(args)

        # --- run the task ---
        self.task.instantiate_targets()
        self.run_task()

        # --- get results ---
        roc_dir_list = self.task.outputs['roc_dir_list'].data
        roc_data_files = self.task.outputs['roc_data_files'].data

        # --- make assertions
        # --- DIR LIST ---
        roc_dir_list_all = ['ck', 'fk', 'ik', 'lsk', 'pck', 'sclk', 'spk']
        assert roc_dir_list == roc_dir_list_all, \
            'dir list is not the one expected'

        # --- FILES LIST ---
        # read expected data
        expected_data = DingoLogFileTest.open_test_data(
            expected_data_file, root)

        # compare resulsts with expected data
        DingoLogFileTest.compare_data_to_expected(
            expected_data, roc_data_files)



# -*- coding: utf-8 -*-
"""empty message

Revision ID: roc_dingo_0007_add_error_log
Revises: roc_dingo_0006_add_is_removed
Create Date: 2020-12-17 11:38:09.315512

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'roc_dingo_0007_add_error_log'
down_revision = 'roc_dingo_0006_add_is_removed'
branch_labels = None
depends_on = None


def upgrade():

    # ADD error_log column

    op.add_column(
        'file_log',
        sa.Column('error_log', sa.String(), nullable=True),
        schema='pipeline')

    # ADD SPICE_KERNEL in the levels

    op.execute('COMMIT')
    op.execute("ALTER TYPE file_level_list ADD VALUE IF NOT EXISTS 'SK'")


def downgrade():
    op.drop_column('file_log', 'error_log', schema='pipeline')

    # Postgresl does not offer the ability to remove a value from an
    # enumerate type
    # Sk is not removed from enumerate type
    # the 'IF EXISTS' option is set in upgrade() to allow
    # multiple succesive upgrade/downgrade



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Tests module for the roc.dingo plugin.
"""
import pytest
import os.path as osp
from roc.dingo.models.file import FileLog
from roc.dingo.tests.test_log_file import \
    DingoLogFileCommandTestCase, DingoLogFileTest


# Tests on roc.dingo.tasks.log_file methods
class TestDingoLogFileToRocdb(DingoLogFileCommandTestCase):

    def setup_method(self, method):
        super().setup_method(method)

        # empty the database
        self.session.query(FileLog).delete()
        self.session.flush()

    def teardown_method(self, method):
        super().teardown_method(method)

        # close database
        self.session.close()

    @pytest.mark.parametrize('root, date, level, nb_files', [
        ('DATA-1', None, None, 61),
        ('DATA-1', '2020/08/19', None, 31),
        ('DATA-1', '2020/08/19', ['L0'], 1),
        ('DATA-1', '2020/08/19', ['L0', 'L1'], 14),
        ('DATA-1', '2020/08/19', ['L2'], 0),
        ('DATA-1', '2020/08', ['L2'], 17),
    ])
    def test_logfile_to_rocdb(self, root, date, level, nb_files):
        """
        Test insertion with different input parameters
        Check the number of files inserted in db after each
        """
        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)

        # --- initialize the command ---
        main_command = ['pop',
                        'dingo',
                        'logfile_to_rocdb'
                        ]

        if root is not None:
            main_command += ['--root', root]

        if date is not None:
            main_command += ['--date', date]

        if level is not None:
            main_command += ['--level'] + level

        # --- run the command ---
        self.run_command(main_command)

        # --- make assertions ---

        results = self.session.query(FileLog).all()
        assert len(results) == nb_files

        self.session.close()

    @pytest.mark.parametrize('root', [
        ('DATA-1'),
    ])
    def test_logfile_to_rocdb_multiple(self, root):
        """
        Test multiple insertion with different input parameters
        Check the number of files inserted in db after each
        """

        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)

        # --- initialize the commands ---
        commands = [['pop',
                     'dingo',
                     'logfile_to_rocdb',
                     '--root', root,
                     '--date', '2020/08/19',
                     '--level', 'L0', 'L1'],
                    ['pop',
                     'dingo',
                     'logfile_to_rocdb',
                     '--root', root,
                     '--date', '2020/08/19',
                     '--level', 'L0'],
                    ['pop',
                     'dingo',
                     'logfile_to_rocdb',
                     '--root', root,
                     '--date', '2020/08',
                     '--level', 'L2']
                    ]

        nb_files = [14, 14, 31]

        for i, cmd in enumerate(commands):
            # --- run the  command ---
            self.run_command(cmd)

            # --- make assertions ---
            results = self.session.query(FileLog).all()
            assert len(results) == nb_files[i]

        self.session.close()

    @pytest.mark.parametrize('root', [
        ('SK-1'),
    ])
    def test_logsk_to_rocdb(self, root):
        """
        Test spice kernel insertion
        Check the number of files inserted in db after each
        """
        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        root = osp.join(data_test_path, root)

        # --- initialize the commands ---
        command = ['pop',
                   'dingo',
                   'logsk_to_rocdb',
                   '--root', root
                   ]

        nb_files = 5

        # --- run the  command ---
        self.run_command(command)

        # --- make assertions ---
        results = self.session.query(FileLog).filter(
            FileLog.level == 'SK').all()
        assert len(results) == nb_files

        self.session.close()

    @pytest.mark.parametrize(
        'filesroot, skroot, expected_before, expected_after',
        [(
            'DATA-1', 'SK-1', 32, 12
        )]
    )
    def test_update_missing_parents_in_db(
            self, filesroot, skroot, expected_before, expected_after):
        """
        Test spice kernel insertion
        Check the number of files inserted in db after each
        """

        # --- set up arguments ---
        data_test_path = DingoLogFileTest.get_test_data_path()
        filesroot = osp.join(data_test_path, filesroot)
        skroot = osp.join(data_test_path, skroot)

        command = ['pop',
                   'dingo',
                   'logfile_to_rocdb',
                   '--root', filesroot
                   ]
        self.run_command(command)
        command = ['pop',
                   'dingo',
                   'logsk_to_rocdb',
                   '--root', skroot
                   ]
        self.run_command(command)

        results = self.session.query(FileLog).filter(
            FileLog.error_log.like('%Missing parents%')).all()
        assert len(results) == expected_before

        command = ['pop',
                   'dingo',
                   'update_missing_parents_in_db',
                   '--root', filesroot
                   ]
        self.run_command(command)

        results = self.session.query(FileLog).filter(
            FileLog.error_log.like('%Missing parents%')).all()
        assert len(results) == expected_after

        self.session.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Functions used in all DINGO FileLog tests
"""
import os
import re
from datetime import datetime

from poppy.core.logger import logger
from poppy.core.test import TaskTestCase, CommandTestCase

from roc.dingo.models.file import FileLog
from roc.dingo.constants import \
    DATA_ALLOWED_EXTENSIONS, \
    SPICE_KERNEL_ALLOWED_EXTENSIONS
from roc.dingo.tests.test_dingo import DingoTest


class DingoLogFileTaskTestCase(TaskTestCase):

    @classmethod
    def setup_class(cls):
        logger.debug('setup_class DingoLogFileTaskTestCase')
        d = DingoLogFileTest()
        d.get_test_data()

    #     # --- database setup ---
    #     session = DingoTestLogFile.setup_session()

    # @classmethod
    # def teardown_class(cls):
    #     logger.debug('teardown_class DingoTaskTestCase')

    #     # --- close session
    #     session.close()


class DingoLogFileCommandTestCase(CommandTestCase):

    @classmethod
    def setup_class(cls):
        logger.debug('setup_class DingoLogFileCommandTestCase')
        d = DingoLogFileTest()
        d.get_test_data()


class DingoLogFileTest(DingoTest):

    base_url = f'{DingoTest.base_url}/logfile_to_rocdb'
    base_path = f'{DingoTest.base_path}/logfile_to_rocdb'

    def open_test_data(expected_filename, root):
        """
        Open <expected_filename> file, which contains a list of
            file1 size1 timestamp1
            file2 size2 timestamp2
            ...
        And creates a dictionnary array

            :param expected_filename: filename containing expected data
            :param root: the data root in the file system
            :return: a dictionnary array
        """

        # open the data_files file with data for test
        path = os.path.abspath(os.path.dirname(__file__))
        path = os.path.join(path, 'data-tasks', expected_filename)
        expected_data = {}
        with open(path, 'r') as f:
            # read a line corresponding to a file
            for line in f.readlines():
                relpath, size, cdate = line.split()
                basename = os.path.basename(relpath)
                # append the data
                expected_data[basename] = {
                    'filepath': os.path.join(
                        root, relpath),
                    'size': int(size),
                    'creation_time': datetime.utcfromtimestamp(int(cdate))
                }

        return expected_data

    def data_to_file_log(data, root):
        dirname = os.path.dirname(data['filepath']).replace(root, '')
        basename = os.path.basename(data['filepath'])
        file_path = dirname.split('/')
        file_name, file_extension = os.path.splitext(basename)

        is_data = True
        re_ext = '|'.join(DATA_ALLOWED_EXTENSIONS)

        if file_extension.replace('.', '') in SPICE_KERNEL_ALLOWED_EXTENSIONS:
            is_data = False
            re_ext = '|'.join(SPICE_KERNEL_ALLOWED_EXTENSIONS)

        if is_data:
            level = file_path[0]
            res = re.search(r'(v|V)([0-9a-zA-Z]+)\.(' + re_ext + ')', basename)
            res_id = 2
        else:
            level = 'SK'
            res = re.search(r'([0-9]+)\.(' + re_ext + ')', basename)
            res_id = 1

        try:
            version = res.group(res_id)
        except Exception:
            version = '0'

        product = re.sub(r'^([^\.]+)_V[0-9U]+\.[^$]+$', r'\1', basename)

        file_log = {
            'basename': basename,
            'product': product,
            'dirname': dirname,
            'size': data['size'],
            'creation_time': data['creation_time'],
            'version': version,
            'state': 'OK',
            'status': 'Terminated',
            'level': level,
            'insert_time': datetime.now(),
            'is_archived': False,
            'is_latest': True
        }
        return FileLog(**file_log)

    def setup_session():

        session = DingoTest.setup_session()

        # empty the table FileLog
        session.query(FileLog).delete()
        session.flush()

        return session



# -*- coding: utf-8 -*-
"""Add a columns to store if file has already been delivered to ESA
and with which filename

Revision ID: roc_dingo_0016_add_is_delivered
Revises: roc_dingo_0015_add_dataqueue
Create Date: 2021-10-12 11:58:35.441300

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'roc_dingo_0018a_task_delivered'
down_revision = 'roc_dingo_0017_add_efecsevents'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        'file_log',
        sa.Column('is_delivered', sa.BOOLEAN(), nullable=True),
        schema='pipeline')

    op.add_column(
        'file_log',
        sa.Column('public_filename', sa.String(), nullable=True),
        schema='pipeline')

    op.add_column(
        'file_log',
        sa.Column('public_dirname', sa.String(), nullable=True),
        schema='pipeline')

    # adding a column with a default value is not possible with Alembic
    # so create the column without the nullable constraint
    # fill the table with the default value (false)
    # after, add the nullable constraint

    op.execute('UPDATE pipeline.file_log SET is_delivered = false')
    op.alter_column(
        'file_log', 'is_delivered',
        nullable=False, schema='pipeline')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    op.drop_column('file_log', 'is_delivered', schema='pipeline')
    op.drop_column('file_log', 'public_filename', schema='pipeline')
    op.drop_column('file_log', 'public_dirname', schema='pipeline')
    # ### end Alembic commands ###



"""empty message

Revision ID: roc_dingo_0016_add_kcoeff_biascurr
Revises: roc_dingo_0015_add_dataqueue
Create Date: 2021-10-10 12:40:44.668707

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

from poppy.pop.alembic.helpers import \
    create_table, create_schema, execute, user, drop_table

# revision identifiers, used by Alembic.
revision = 'roc_dingo_0016_add_kcoeffbiacur'
down_revision = 'roc_dingo_0015_add_processqueue'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    create_table('bia_current_log',
                 sa.Column('id', sa.BIGINT(), nullable=False),
                 sa.Column('tc_name', sa.String(length=256), nullable=False),
                 sa.Column('utc_time', postgresql.TIMESTAMP(), nullable=False),
                 sa.Column('bia_current', sa.FLOAT(), nullable=False),
                 sa.Column('insert_time', postgresql.TIMESTAMP(),
                           nullable=True),
                 sa.PrimaryKeyConstraint('id'),
                 sa.UniqueConstraint('utc_time', 'tc_name'),
                 schema='pipeline'
                 )
    create_table('lfr_kcoeff_dump',
                 sa.Column('id', sa.BIGINT(), nullable=False),
                 sa.Column('utc_time', postgresql.TIMESTAMP(), nullable=False),
                 sa.Column('cuc_time', sa.String(length=1024), nullable=False),
                 sa.Column('utc_time_is_predictive',
                           sa.BOOLEAN(), nullable=True),
                 sa.Column('kcoeff_pkt_cnt', sa.SMALLINT(), nullable=False),
                 sa.Column('kcoeff_pkt_nr', sa.SMALLINT(), nullable=False),
                 sa.Column('kcoeff_blk_nr', sa.SMALLINT(), nullable=False),
                 sa.Column('kcoeff_values', postgresql.JSONB(astext_type=sa.Text()), nullable=False),
                 sa.Column('insert_time', postgresql.TIMESTAMP(),
                           nullable=True),
                 sa.PrimaryKeyConstraint('id'),
                 sa.UniqueConstraint('utc_time', 'kcoeff_pkt_nr'),
                 schema='pipeline'
                 )
    op.alter_column('solo_hk_param', 'description',
                    existing_type=sa.VARCHAR(),
                    nullable=False,
                    schema='pipeline')
    op.alter_column('solo_hk_param', 'eng_value',
                    existing_type=sa.VARCHAR(),
                    nullable=False,
                    schema='pipeline')
    op.alter_column('solo_hk_param', 'raw_value',
                    existing_type=sa.VARCHAR(),
                    nullable=False,
                    schema='pipeline')
    op.alter_column('solo_hk_param', 'unit',
                    existing_type=sa.VARCHAR(),
                    nullable=False,
                    schema='pipeline')
    # ### end Alembic commands ###

    # added by hand
    # fix table permissions
    grant_user_seq = (
        """GRANT USAGE, SELECT
        ON ALL SEQUENCES IN SCHEMA pipeline TO {0}""".format(user)
    )
    execute(grant_user_seq)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('solo_hk_param', 'unit',
                    existing_type=sa.VARCHAR(),
                    nullable=True,
                    schema='pipeline')
    op.alter_column('solo_hk_param', 'raw_value',
                    existing_type=sa.VARCHAR(),
                    nullable=True,
                    schema='pipeline')
    op.alter_column('solo_hk_param', 'eng_value',
                    existing_type=sa.VARCHAR(),
                    nullable=True,
                    schema='pipeline')
    op.alter_column('solo_hk_param', 'description',
                    existing_type=sa.VARCHAR(),
                    nullable=True,
                    schema='pipeline')
    drop_table('lfr_kcoeff_dump', schema='pipeline')
    drop_table('bia_current_log', schema='pipeline')
    # ### end Alembic commands ###



"""empty message

Revision ID: roc_dingo_0014_add_solohk
Revises: roc_dingo_0013_fix_db
Create Date: 2021-09-14 16:41:50.725120

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

from poppy.pop.alembic.helpers import \
    create_table, create_schema, execute, user, drop_table

# revision identifiers, used by Alembic.
revision = 'roc_dingo_0014_add_solohk'
down_revision = 'roc_dingo_0013_fix_db'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    create_table('solo_hk_param',
    sa.Column('id', sa.BIGINT(), nullable=False),
    sa.Column('name', sa.String(length=16), nullable=False),
    sa.Column('utc_time', postgresql.TIMESTAMP(), nullable=False),
    sa.Column('unit', sa.String(), nullable=True),
    sa.Column('description', sa.String(), nullable=True),
    sa.Column('eng_value', sa.String(), nullable=True),
    sa.Column('raw_value', sa.String(), nullable=True),
    sa.Column('sha', sa.String(), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('sha'),
    schema='pipeline'
    )
    # ### end Alembic commands ###

    # added by hand
    # fix table permissions
    grant_user_seq = (
        """GRANT USAGE, SELECT
        ON ALL SEQUENCES IN SCHEMA pipeline TO {0}""".format(user)
    )
    execute(grant_user_seq)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    drop_table('solo_hk_param', schema='pipeline')
    # ### end Alembic commands ###



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Functions used in all DINGO tests
"""
import os
import os.path as osp
import copy

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from poppy.core.logger import logger
from poppy.core.generic.requests import download_file

from roc.dingo.constants import TEST_DATABASE

# Models to be adapted for each Dingo model
#
# from poppy.core.test import TaskTestCase, CommandTestCase
# class DingoTaskTestCase(TaskTestCase):
#
#     @classmethod
#     def setup_class(cls):
#         logger.debug('setup_class DingoLogFileTaskTestCase')
#         d = DingoTest()
#         d.get_test_data()
#
#    #     # --- database setup ---
#    #     session = DingoTestLogFile.setup_session()
#
#    # @classmethod
#    # def teardown_class(cls):
#    #     logger.debug('teardown_class DingoTaskTestCase')
#
#    #     # --- close session
#    #     session.close()
#
# class DingoCommandTestCase(CommandTestCase):
#
#     @classmethod
#     def setup_class(cls):
#         logger.debug('setup_class DingoCommandTestCase')
#         d = DingoTest()
#         d.get_test_data()


class DingoTest():

    base_url = 'https://rpw.lesia.obspm.fr/roc/data/private/devtest/roc/test_data/rodp/dingo'  # noqa: E501
    base_path = '/volumes/plasma/rpw/roc/data/https/private/devtest/roc/test_data/rodp/dingo'  # noqa: E501

    # test credentials
    host = 'roc2-dev.obspm.fr'
    username = 'roctest'
    password = None

    def __init__(self):
        logger.debug('DingoTest setup_class()')
        logger.debug(f'base_url = {self.base_url}')
        logger.debug(f'base_path = {self.base_path}')
        try:
            self.password = os.environ['ROC_TEST_PASSWORD']
        except KeyError:
            raise KeyError('You have to define the test user password using'
                           'the "ROC_TEST_PASSWORD" environment variable')

    def compare_data_to_expected(expected_data, data):
        """
        Compare expected_data to the data array dictionnary
            -> check if every expected item is in data
            -> check if every data item is in expected data

            :param expected_data: expected data
            :param data: the data to be compared
        """
        # check length
        assert len(expected_data) == len(data)

        # check content
        for item in data:
            assert item in expected_data, \
                'item {} is not in expected data'.format(item)

        for item in expected_data:
            assert item in data, \
                'item {} is not in result'.format(item)
            assert data[item] == expected_data[item], \
                'item {} is not the one expected :' \
                'expected {} / got {}'.format(
                    item, expected_data[item], data[item])

    def read_data(file):
        """
        Read a file located in the data-tasks directory
        and returns file content as an array of strings

            :param file: file to open
        """
        path = os.path.abspath(os.path.dirname(__file__))
        path = os.path.join(path, 'data-tasks', file)
        data = []
        with open(path, 'r') as f:
            # read a line corresponding to a directory
            for line in f.readlines():
                # append the data
                data.append(line.strip())
        return data

    def get_test_data_path():
        """
        Read the config file and returns ROC_DINGO_TEST_DATA_PATH
        which is the path where to store the test dataset locally
        """
        conf = DingoTest.load_configuration()

        try:
            data_test_path = conf['environment']['ROC_DINGO_TEST_DATA_PATH']
        except Exception:
            logger.info('Env. variable ROC_DINGO_TEST_DATA_PATH not set')
            logger.info('Using /tmp/DINGO_TEST_DATA/')
            data_test_path = '/tmp/DINGO_TEST_DATA/'

        logger.debug(f'ROC_DINGO_TEST_DATA_PATH = {data_test_path}')
        return data_test_path

    def get_test_data(self):
        """
        Get the test dataset indicated by the environment variable
        ROC_DINGO_TEST_DATA_PATH

        :param subdir: optional subdirectory

        Try to make a rsync with the roctest account
        A public kay has to be setup on the server to allow connexion
        If the command is not available (Windows),
        use the download_file() method
        """
        data_test_path = DingoTest.get_test_data_path()
        os.makedirs(data_test_path, exist_ok=True)

        try:
            logger.info('Starting rsync')
            ssh_option = '\"ssh -o \'StrictHostKeyChecking no\'\"'
            rsync_cmd = 'rsync -e {} -irtzuv {}@{}:{}/ {}/'.format(
                ssh_option,
                self.username,
                self.host,
                self.base_path,
                data_test_path
            )
            logger.info('Executing ' + rsync_cmd)
            output = os.popen(rsync_cmd)
            rsync_output = output.read()
            if output.close() is not None:
                raise ValueError('Rsync failed : {}'.format(rsync_output))
        except ValueError:
            logger.info('Rsync failed, using download_test_data()')
            self.download_test_data(data_test_path)

    def download_test_data(self, data_test_path):
        """
        Download the manifest.txt file located at self.base_url
        And for each file, download it only if the file does not exist
        in data_test_path
        """
        logger.debug('download_test_data()')
        manifest_filepath = osp.join(data_test_path, 'manifest.txt')

        manifest_file_url = self.base_url + '/manifest.txt'
        auth = (self.username, self.password)

        file_list = list(self.load_manifest_file(
            manifest_filepath, manifest_file_url, auth=auth))

        for relative_filepath in file_list:
            # skip empty strings
            if not relative_filepath:
                continue

            # get the complete filepath
            filepath = osp.join(data_test_path, relative_filepath)
            os.makedirs(osp.dirname(filepath), exist_ok=True)

            # download it only if it does not exist
            if not osp.isfile(filepath):
                logger.info('Downloading {}'.format(filepath))
                download_file(filepath,
                              f'{self.base_url}/{relative_filepath}',
                              auth=auth)

    def load_manifest_file(
            self, manifest_filepath, manifest_file_url, auth=None):
        """
        Read the manifest.txt file located at manifest_file_url
        and returns the list composed by the file list
        """

        download_file(manifest_filepath, manifest_file_url, auth=auth)

        with open(manifest_filepath) as manifest_file:
            for line in manifest_file:
                yield line.strip('\n\r')

        os.remove(manifest_filepath)

    def load_configuration():
        from poppy.core.configuration import Configuration

        configuration = Configuration(os.getenv('PIPELINE_CONFIG_FILE', None))
        configuration.read()

        return configuration

    def setup_session():
        # Read config file
        conf = DingoTest.load_configuration()

        database_info = list(filter(
            lambda db: db['identifier'] == TEST_DATABASE,
            conf['pipeline.databases']))[0]

        # Create an Engine, which the Session will use for connection resources
        engine = create_engine('{}://{}@{}/{}'.format(
            database_info['login_info']['vendor'],
            database_info['login_info']['user'],
            database_info['login_info']['address'],
            database_info['login_info']['database']
        ))
        # create a configured "Session" class
        Session = sessionmaker(bind=engine, autocommit=False)
        # create a Session
        session = Session()

        return session

    def get_db_values(item):
        """
        returns a dict for any object in DB returned by session.query

        :param item: the object
        """

        return copy.copy(item.__dict__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to synchronize the file tree with the ROC database."""

import os
import re

from datetime import datetime, timedelta

from sqlalchemy import and_
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import load_only

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task
from poppy.core.target import BaseTarget

from roc.dingo.constants import PIPELINE_DATABASE, ROC_DATA_ROOT, \
    DATA_ALLOWED_EXTENSIONS, SPICE_KERNEL_ALLOWED_EXTENSIONS, \
    ROC_DATA_HTTPS_ROOT, ROC_SPICE_KERNEL_HTTPS_ROOT
from roc.dingo.models.file import FileLog

from roc.film.constants import TIME_ISO_STRFORMAT

import h5py
from spacepy.pycdf import CDF
from astropy.time import Time


__all__ = ['LogFileToDb']


class LogFileToDb(Task):
    """
    Parse ROC file tree and synchronize withe the ROC databaser
    """
    plugin_name = 'roc.dingo'
    name = 'log_file_to_db'
    files_to_update = []
    files_to_insert = []

    def add_targets(self):
        logger.debug('LogFileToDb() : add_targets')
        self.add_input(target_class=BaseTarget, many=True,
                       identifier='roc_data_files_to_insert')
        self.add_input(target_class=BaseTarget, many=True,
                       identifier='roc_data_files_to_update')

    def setup_inputs(self):
        """
        Setup task inputs.

        :param task:
        :return:
        """

        logger.debug('LogFileToDb() : setup_inputs')

        # get the root file tree
        self.root = self.pipeline.get('root', default=ROC_DATA_ROOT, args=True)
        # ensure that there is / at the end
        self.root = os.path.join(self.root, '')

        # get files to update
        self.files_to_db = {}
        self.files_to_db['update'] = self.inputs['roc_data_files_to_update']
        self.files_to_db['insert'] = self.inputs['roc_data_files_to_insert']

    def run(self):

        logger.debug('LogFileToDb() : run')

        # Get the database connection if needed
        if not hasattr(self, 'session'):
            self.session = Connector.manager[PIPELINE_DATABASE].session

        # Initialize inputs
        self.setup_inputs()

        # Insert / Update files
        for action, files in self.files_to_db.items():
            if files.is_empty:
                continue

            logger.debug('*** TO {} ***'.format(action.upper()))
            num_files = len(files.data)
            cpt_file_ok = 0
            cpt_file_ko = 0
            for basename, item in files.data.items():
                logger.debug(item['filepath'])

                try:
                    file_dict = LogFileToDb.item_to_file_log(
                        item, self.root)
                except ValueError as e:
                    logger.error('Value error:', e)
                    if action == 'update':
                        # We only store that there was an error
                        file_dict = {
                            'basename': basename,
                            'status': 'Failed'}
                    else:
                        cpt_file_ko += 1
                        continue
                except Exception as e:
                    cpt_file_ko += 1
                    logger.error('Unexpected error:', e)
                    continue

                # retreive parents ID (1/2)
                # remove "parents" item before inserting
                parents = file_dict.pop('parents')

                try:
                    if action == 'insert':
                        file_log = FileLog(**file_dict)
                        query = self.session.add(file_log)
                    else:
                        # action == 'update'
                        query = self.session.query(FileLog).\
                            filter(FileLog.basename == basename).\
                            update(file_dict)
                        file_log = self.session.query(FileLog).\
                            filter(FileLog.basename == basename).first()
                except Exception:
                    cpt_file_ko += 1
                    logger.error('{} file {} has failed: \n {!s}'.format(
                        action.capitalize,
                        file_dict['basename'],
                        query.statement.compile(
                            dialect=postgresql.dialect(),
                            compile_kwargs={'literal_binds': True})
                    ))
                    # break to the following file
                    continue
                else:
                    cpt_file_ok += 1
                    logger.info(
                        '{} file {:60s} OK [{:05.2f}% complete]'.format(
                            action.capitalize(),
                            file_dict['basename'],
                            100 * cpt_file_ok / num_files
                        ))

                # ensure that previous versions are set to is_latest=False
                # get file filename without extension and version
                if file_dict['level'] != 'SK':
                    trunc_basename = os.path.splitext(
                        file_dict['basename'])[0].\
                        replace('_V' + file_dict['version'], '')
                    logger.debug(f'TRUNC : {trunc_basename}')

                    base_filter = FileLog.basename.like(
                        '%' + trunc_basename + '%')
                    latest_filter = FileLog.is_latest == True  # noqa: E712
                    version_filter = FileLog.version < file_dict['version']

                    try:
                        query = self.session.query(FileLog).\
                            filter(and_(
                                base_filter,
                                latest_filter,
                                version_filter))

                        for r in query.all():
                            logger.debug(f'OLD VERSION : {r.basename}')

                        old_items = {}
                        old_items['is_latest'] = False
                        nb = query.update(old_items, synchronize_session=False)

                    except Exception as e:
                        logger.error(
                            f'Error while searching previous versions : {e}')

                    else:
                        if nb > 0:
                            logger.info(
                                f'Setting {nb} files to is_latest = False')

                # retreive parents ID (2/2)
                parents_added = []
                parents_filter = FileLog.basename.in_(parents)
                logger.debug('Parents : {}'.format(parents))
                try:
                    query = self.session.query(FileLog).\
                        options(load_only('id')).\
                        filter(parents_filter)
                except Exception:
                    logger.error('{} file {} has failed: \n {!s}'.format(
                        action.capitalize,
                        file_dict['basename'],
                        query.statement.compile(
                            dialect=postgresql.dialect(),
                            compile_kwargs={'literal_binds': True})
                    ))
                else:
                    for result in query.all():
                        file_log.parents.append(result)
                        parents_added.append(result.basename)
                        parents.remove(result.basename)

                        # sometimes a parent without extension is specified
                        # both extensions cdf/h5 are then added
                        # in the parents list
                        # we have to remove the one which wasn't the good one
                        if '.cdf' in result.basename:
                            h5name = result.basename.replace(
                                '.cdf', '.h5')
                            if h5name in parents:
                                parents.remove(h5name)

                        if '.h5' in result.basename:
                            cdfname = result.basename.replace(
                                '.h5', '.cdf')
                            if cdfname in parents:
                                parents.remove(cdfname)

                    # some parents are missing
                    if len(parents) > 0:
                        if file_log.error_log is None:
                            file_log.error_log = ''

                        if file_log.error_log != '':
                            file_log.error_log += '; '

                        file_log.error_log += 'Missing parents : ' + \
                            ', '.join(parents)
                        logger.debug('Missing parents : {}'.format(
                            ', '.join(parents)))

                    self.session.commit()

                logger.debug('Parents added : {}'.format(parents_added))

            logger.info('{} : {} success / {} failure'.format(
                action.capitalize(), cpt_file_ok, cpt_file_ko
            ))

    @staticmethod
    def item_to_file_log(item, root):
        """
        Create a dictionnary ready to be inserted as a FileLog

        :param item: dictionnary with file-system elements already filled :
            size, creation_time, filepath, id
        :param root: the root name of the files (root os not striord in DB)
        :return: dictionnary ready to be inserted as a FileLog
        """

        # DATA or SPICE_KERNEL ?
        is_data = True
        url_prefix = ROC_DATA_HTTPS_ROOT
        re_ext = '|'.join(DATA_ALLOWED_EXTENSIONS)

        file_name, file_extension = os.path.splitext(item['filepath'])
        if file_extension.replace('.', '') in SPICE_KERNEL_ALLOWED_EXTENSIONS:
            url_prefix = ROC_SPICE_KERNEL_HTTPS_ROOT
            is_data = False
            re_ext = '|'.join(SPICE_KERNEL_ALLOWED_EXTENSIONS)

        error_log = []
        warning_log = []

        file_dict = {}
        file_dict['creation_time'] = item['creation_time']
        file_dict['size'] = item['size']

        file_dict['dirname'] = os.path.dirname(
            item['filepath']).replace(root, '')

        file_dict['basename'] = os.path.basename(
            item['filepath'])

        file_dict['sha'] = ''
        file_dict['state'] = 'OK'
        file_dict['status'] = 'Terminated'
        file_dict['insert_time'] = datetime.now()
        file_dict['descr'] = ''
        file_dict['author'] = ''

        file_dict['is_removed'] = not os.path.exists(item['filepath'])
        # Removed file cannot be the latest one
        file_dict['is_latest'] = (not file_dict['is_removed']) and \
            ('former_version' not in file_dict['dirname'])

        file_dict['url'] = url_prefix + '/' + file_dict['dirname'] + '/' + \
            file_dict['basename']

        # Things in path
        file_path = file_dict['dirname'].split('/')
        if is_data:
            file_dict['level'] = file_path[0]
        else:
            file_dict['level'] = 'SK'

        # Things in filename
        if not is_data:
            res = re.search(
                r'^([^\.]+)_V([0-9]+)\.(' + re_ext + ')',
                file_dict['basename'])
        else:
            res = re.search(
                r'^([^\.]+)_(v|V)([0-9a-zA-Z]+)\.(' + re_ext + ')',
                file_dict['basename'])

        if res:
            file_dict['product'] = res.group(1)
        else:
            file_dict['product'] = ''
            logger.error('Unable to get product name in {}'.format(
                file_dict['basename']))
            error_log.append('Unable to get product name in {}'.format(
                file_dict['basename']))

        file_attrs = {
            'TIME_MIN': None,
            'TIME_MAX': None,
            'Dataset_ID': None,
            'Data_version': '00',
            'Parents': [],
            'SPICE_KERNELS': [],
        }

        try:
            if '.h5' in file_dict['basename']:
                file_attrs = LogFileToDb.get_attrs_from_h5(
                    item['filepath'], file_attrs)

            if '.cdf' in file_dict['basename']:
                file_attrs = LogFileToDb.get_attrs_from_cdf(
                    item['filepath'], file_attrs)

        except Exception as e:
            logger.error('Error while reading {} : {}'.format(
                file_dict['basename'], e))
            # import traceback
            # traceback.print_exc()
            error_log.append(str(e))

        parents = file_attrs['Parents']
        start_time = file_attrs['TIME_MIN']
        end_time = file_attrs['TIME_MAX']
        data_set_id = file_attrs['Dataset_ID']
        version = file_attrs['Data_version']

        # fields error_log and warning_log may have been
        # filled during get_attrs_from_xxx()
        if 'error_log' in file_attrs:
            error_log += file_attrs['error_log']

        if 'warning_log' in file_attrs:
            warning_log += file_attrs['warning_log']

        # Version checks
        if is_data and file_extension != '.xml':
            # remove leading 'V'
            res = re.search(r'^[v|V]([0-9a-zA-Z]+)$', version)
            if res:
                logger.debug(f'Removing leading V in version : {version}')
                version = res.group(1)

            # check if the one in filename is the one in attributes
            res = re.search(
                r'(v|V)([0-9a-zA-Z]+)\.(' + re_ext + ')',
                file_dict['basename'])
            if version != '' and res:
                version_in_filename = res.group(2)
                if version != version_in_filename:
                    msg = 'Versions in filename and attributes do not match'
                    logger.error(msg)
                    error_log.append(msg)
            elif version == '' and res:
                version = res.group(2)
                msg = 'No version in attributes'
                logger.warning(msg)
                warning_log.append(msg)
            else:
                version = '00'
                msg = 'No version in filename'
                logger.warning(msg)
                warning_log.append(msg)
        else:
            res = re.search(
                r'_(v|V)([0-9]+)\.(' + re_ext + ')',
                file_dict['basename'])
            if res:
                version = res.group(2)
            else:
                version = '00'

        logger.debug(f'Version =  {version}')

        # Skipping parents[] search for spice kernels
        if is_data:
            logger.debug('Reading regular data')
            # some parents entries are not well formed
            to_be_removed = []
            for i, p in enumerate(parents):
                # Some entries are empty
                if re.match(r'\s*$', p):
                    to_be_removed.append(p)
                    continue

                # Some parents have no file extension
                if not re.search(r'\.(\w{,3})$', p):
                    to_be_removed.append(p)
                    # either extensions can be the good one
                    parents.append(p + '.cdf')
                    parents.append(p + '.h5')

                # ignoring mk/*.tm spice kernels (meta kernel)
                if re.search(r'\.tm$', p):
                    to_be_removed.append(p)

            # finally cleaning parents array
            for i, p in enumerate(to_be_removed):
                parents.remove(p)

        # Reading spice kernels
        if not is_data:
            logger.debug('Reading spice kernels')

            start_time = LogFileToDb.commnt_get_meta(
                item['filepath'], 'START_TIME')
            if len(start_time) > 0:
                start_time = min(start_time)
            else:
                start_time = None

            end_time = LogFileToDb.commnt_get_meta(
                item['filepath'], 'STOP_TIME')
            if len(end_time) > 0:
                end_time = min(end_time)
            else:
                end_time = None

        # When start/stop time have not been set
        # Guess them from basename
        datetimes_expr = \
            re.compile(r'_([0-9]{8}T[0-9]{6})\-([0-9]{8}T[0-9]{6})_')
        dates_expr = re.compile(r'_([0-9]{8})\-([0-9]{8})_')
        date_expr = re.compile(r'_([0-9]{8})_')
        if start_time is None or end_time is None:
            res_times = re.search(datetimes_expr, file_dict['basename'])
            res_dates = re.search(dates_expr, file_dict['basename'])
            res_date = re.search(date_expr, file_dict['basename'])

            if res_times:
                start_time = res_times.group(1)
                end_time = res_times.group(2)

            elif res_dates:
                start_time = res_dates.group(1)
                end_time = res_dates.group(2)

            elif res_date:
                start_time = res_date.group(1)
                try:
                    end_time = (
                        Time.strptime(
                            res_date.group(1), '%Y%m%d') + timedelta(days=1)
                    ).iso
                except ValueError as e:
                    logger.warning(
                        f'Unable to parse end time {end_time} : {e}')
                    end_time = None

            if start_time is not None or end_time is not None:
                logger.info(
                    'Guessing start/end time for {} : {} - {}'.format(
                        file_dict['basename'],
                        start_time,
                        end_time
                    ))

        if len(error_log) > 0:
            file_dict['state'] = 'ERROR'

        elif len(warning_log) > 0:
            file_dict['state'] = 'WARNING'

        # Save both error and warning messages
        if (len(error_log) + len(warning_log)) > 0:
            file_dict['error_log'] = '; '.join(error_log + warning_log)
        else:
            # necessary in order to reset error_log when the error is cleared
            file_dict['error_log'] = ''

        # trimming whitespaces
        file_dict['parents'] = [p.strip() for p in parents]

        file_dict['dataset_id'] = data_set_id
        file_dict['version'] = version
        file_dict['start_time'] = start_time
        file_dict['end_time'] = end_time
        file_dict['validity_end'] = None
        file_dict['validity_start'] = None

        # do not set these values here
        # either they are set to false at creation (table default value)
        # or they are set to True with specific tasks
        # file_dict['is_archived'] = False
        # file_dict['is_delivered'] = False

        file_dict['to_update'] = False

        logger.debug(file_dict)

        return file_dict

    @staticmethod
    def commnt_get_meta(filename, key):
        """
        Retreive parameters in spice kernel using the commnt command

        :param filename: full path of the spice kernel
        :param key: parameter to read
        :return: parameter array values
        """
        val = os.popen(
            'commnt -r ' + filename + ' | '
            "awk -F '=' ' /" + key + "/ {print $2} '").read().split()
        if len(val) == 0:
            if key == 'START_TIME':
                val = LogFileToDb.commnt_get_meta(filename, 'CK-144000START')
            elif key == 'STOP_TIME':
                val = LogFileToDb.commnt_get_meta(filename, 'CK-144000STOP')

            val = [datetime.strptime(
                v, '@%Y-%b-%d-%H:%M:%S.%f').isoformat() for v in val]

        return val

    @staticmethod
    def get_attrs_from_cdf(filename, file_attrs):
        """
        Retreive a attrs from CDF file
            :param filename: full path of the file
            :param file_attrs: dictionnary with the attributes to retreive
            :return: dictionnary file_attrs with the attributes filled plus
                error_log and warning_log
        """

        basename = os.path.basename(filename)
        data_set_id = None
        version = '00'
        parents = []
        parent_version = []
        error_log = []
        warning_log = []

        # Add Parent_version for CDF only
        file_attrs['Parent_version'] = []

        with CDF(filename) as cdf:
            for attr in file_attrs:
                try:
                    logger.debug('{} : {}'.format(basename, attr))
                    # don't forget the [...] to get a copy and not a pointer
                    file_attrs[attr] = cdf.attrs[attr][...]
                except KeyError as e:
                    logger.warning(
                        'Missing attribute in {} : {}'.format(
                            basename,
                            re.sub(r'\'', r'', str(e))))
                    warning_log.append(re.sub(r'\'', r'', str(e)))
                    continue

        try:
            start_time = Time(
                float(file_attrs['TIME_MIN'][0]),
                format='jd', precision=9)
        except ValueError as e:
            start_time = Time('1900-01-01')  # fake date
            msg = 'Formatting date error for time_min in {} : '\
                '{} ({})'.format(
                    basename,
                    re.sub(r'\'', r'', str(e)),
                    file_attrs['TIME_MIN'][0])
            logger.error(msg)
            error_log.append(msg)

        try:
            end_time = Time(
                float(file_attrs['TIME_MAX'][0]),
                format='jd', precision=9)
        except ValueError as e:
            end_time = Time('1900-01-01')  # fake date
            msg = 'Formatting date error for time_max in {} : '\
                '{} ({})'.format(
                    basename,
                    re.sub(r'\'', r'', str(e)),
                    file_attrs['TIME_MAX'][0])
            logger.error(msg)
            error_log.append(msg)

        # Check if data seems to be coherent
        # we sometimes get
        # WARNING: ErfaWarning: ERFA function "d2dtf" yielded
        # 1 of "dubious year (Note 5)" [astropy._erfa.core]

        if start_time.jd < Time('2020-02-11').jd or \
           start_time.jd > Time('2100-01-01').jd:
            msg = f'Start time seems to be erroneous : {start_time.iso}'
            logger.error(msg)
            error_log.append(msg)
            start_time = None
        else:
            start_time = start_time.iso

        if end_time.jd < Time('2020-02-11').jd or \
           end_time.jd > Time('2100-01-01').jd:
            msg = f'End time seems to be erroneous : {end_time.iso}'
            logger.error(msg)
            error_log.append(msg)
            end_time = None
        else:
            end_time = end_time.iso

        try:
            data_set_id = file_attrs['Dataset_ID'][0]
            # Clean data_set_id
            # some files have {} around id
            data_set_id = re.sub(r'[\{\}]', '', data_set_id)
        except Exception:
            logger.warning('Dataset_ID missing')

        try:
            version = f'{file_attrs["Data_version"][0]}'
            # ensure version number is a string on 2 digits
            version = version.strip()
            if len(version) == 1:
                version = f'0{version}'
        except Exception:
            version = '00'
            logger.warning('Data_version missing')

        # some files store one parents in each entry,
        # other in string concatenated with commas
        for entry in file_attrs['Parents']:
            if entry.strip() != '':
                parents += entry.split(',')

        for entry in file_attrs['Parent_version']:
            # ensure it is a string
            entry = f'{entry}'
            if entry.strip() != '':
                parent_version += entry.split(',')

        # remove extra whitespace
        parents = [p.strip() for p in parents]
        parent_version = [p.strip() for p in parent_version]

        logger.debug('*** PARENTS ***')
        logger.debug(f'{parents}')
        logger.debug(f'{parent_version}')

        # if parents and parent_version don't have the same length
        # it is useless to try to append the version number
        if len(parents) != len(parent_version):
            logger.debug(parents)
            logger.debug(parent_version)
            error_log.append('Parents and Parent_version length mismatch')
        else:
            # Append the version number if needed
            for i, p in enumerate(parents):
                p_name, p_ext = os.path.splitext(p)
                if not re.search(r'_V[0-9U]+$', p_name):
                    version = parent_version[i]
                    # some files say '5' instead of '05'
                    if len(version) == 1:
                        version = f'0{version}'
                    p_name += f'_V{version}'
                    parents[i] = p_name + p_ext
                    logger.debug('Adding version to parent name : '
                                 f'{parents[i]}')

        # some files store spice kernels in array,
        # other in string concatenated with commas
        for entry in file_attrs['SPICE_KERNELS']:
            parents += entry.split(',')

        for i, p in enumerate(parents):
            if (p.find('CDF>') != -1):
                parents[i] = p.replace('CDF>', '')
            if (p.find('L0>') != -1):
                parents[i] = p.replace('L0>', '')
            if (p.find('ANC>') != -1):
                parents[i] = p.replace('ANC>', '')

        file_attrs['TIME_MIN'] = start_time
        file_attrs['TIME_MAX'] = end_time
        file_attrs['Dataset_ID'] = data_set_id
        file_attrs['Parents'] = parents
        file_attrs['Data_version'] = version

        file_attrs['error_log'] = error_log
        file_attrs['warning_log'] = warning_log

        return file_attrs

    @staticmethod
    def get_attrs_from_h5(filename, file_attrs):
        """
        Retreive a attrs from HDF5 file
            :param filename: full path of the file
            :param file_attrs: dictionnary with the attributes to retreive
            :return: dictionnary file_attrs with the attributes filled plus
                error_log and warning_log
        """

        basename = os.path.basename(filename)
        error_log = []
        warning_log = []

        with h5py.File(filename, 'r') as l0:
            for attr in file_attrs:
                try:
                    logger.debug('{} : {}'.format(basename, attr))
                    file_attrs[attr] = l0.attrs[attr]
                except KeyError as e:
                    logger.warning('Missing attribute in {} : {}'.format(
                        basename, e))
                    warning_log.append(str(e))
                    continue

        # Get TIME_MIN/TIME_MAX L0 attributes value as datetime
        file_attrs['TIME_MIN'] = datetime.strptime(
            file_attrs['TIME_MIN'], TIME_ISO_STRFORMAT)

        file_attrs['TIME_MAX'] = datetime.strptime(
            file_attrs['TIME_MAX'], TIME_ISO_STRFORMAT)

        file_attrs['Parents'] = file_attrs['Parents'].split(',')
        if len(file_attrs['SPICE_KERNELS']) > 0:
            file_attrs['Parents'] += file_attrs['SPICE_KERNELS'].split(',')

        # Ensure data version is a string on 2 digits
        file_attrs['Data_version'] = f'{file_attrs["Data_version"]}'
        file_attrs['Data_version'] = file_attrs['Data_version'].strip()
        if len(file_attrs['Data_version']) == 1:
            file_attrs['Data_version'] = f'0{file_attrs["Data_version"]}'

        file_attrs['error_log'] = error_log
        file_attrs['warning_log'] = warning_log

        return file_attrs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to synchronize the file tree with the ROC database."""

import os


from sqlalchemy import and_
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import load_only

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task
from poppy.core.target import BaseTarget

from roc.dingo.constants import PIPELINE_DATABASE, ROC_DATA_ROOT
from roc.dingo.models.file import FileLog


__all__ = ['SearchMissingParentsFiles']


class SearchMissingParentsFiles(Task):
    """
    Parse ROC db and search for files with Missing Parents in the error_log
    """
    plugin_name = 'roc.dingo'
    name = 'search_missing_parents_files'
    files_to_update = []

    def add_targets(self):
        logger.debug('SearchMissingParentsFiles() : add_targets')

        self.add_output(target_class=BaseTarget, many=True,
                        identifier='roc_data_files_to_update')

        # useless here but necessary for LogFileToDb()
        self.add_output(target_class=BaseTarget, many=True,
                        identifier='roc_data_files_to_insert')

    def setup_inputs(self):
        """
        Setup task inputs.

        :param task:
        :return:
        """

        logger.debug('SearchMissingParentsFiles() : setup_inputs')

        # get the root file tree
        self.root = self.pipeline.get('root', default=ROC_DATA_ROOT, args=True)
        # ensure that there is / at the end
        self.root = os.path.join(self.root, '')

    def run(self):

        logger.debug('SearchMissingParentsFiles() : run')

        # Get the database connection if needed
        if not hasattr(self, 'session'):
            self.session = Connector.manager[PIPELINE_DATABASE].session

        # Initialize inputs
        self.setup_inputs()

        # Initialize list of ROC data files to update into db
        file_list = {}

        logger.debug('Root directory: {}'.format(self.root))

        # Define relevant fields to retreive
        fields = ['basename', 'dirname', 'creation_time', 'size']

        # Search files with "Missing parents in error_log"
        missing_filter = FileLog.error_log.like('%Missing parents%')
        not_removed_filter = FileLog.is_removed == False  # noqa: E712
        try:
            query = self.session.query(FileLog).\
                options(load_only(*fields)).\
                filter(and_(missing_filter, not_removed_filter))
            logger.debug(
                str(query.statement.compile(
                    dialect=postgresql.dialect(),
                    compile_kwargs={'literal_binds': True})))
            results = query.all()
        except Exception:
            logger.error('Query has failed: \n {!s}'.format(
                query.statement.compile(
                    dialect=postgresql.dialect(),
                    compile_kwargs={'literal_binds': True})
            ))
        else:
            for item in results:
                # adding an entry in file_list
                file_list[item.basename] = {
                    'filepath': os.path.join(
                        self.root, item.dirname, item.basename),
                    'size': item.size,
                    'creation_time': item.creation_time
                }

        # Returns outputs
        self.outputs['roc_data_files_to_update'].data = file_list
        logger.info('{} files to update'.format(str(len(file_list))))
        if len(file_list) == 0:
            self.outputs['roc_data_files_to_update'].is_empty = True

        self.outputs['roc_data_files_to_insert'].data = {}
        self.outputs['roc_data_files_to_insert'].is_empty = True



# -*- coding: utf-8 -*-
"""empty message

Revision ID: roc_dingo_0022_add_to_update_flag_to_filelog
Revises: roc_dingo_0021_add_retrieved_time
Create Date: 2022-01-11 16:07:33.943889

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'roc_dingo_0022_add_to_update_flag_to_filelog'
down_revision = 'roc_dingo_0021_add_retrieved_time'
branch_labels = None
depends_on = None


def upgrade():
    # Add "to_update" flag with a default value to false
    # Trick to set a default value to the non null column
    # https://gist.github.com/VladimirPal/e256d19df8d85234dde1
    # it avoids to create with nullable=true, set the colum to false and
    # then set nullable=false
    op.add_column('file_log',
                  sa.Column('to_update',
                            sa.BOOLEAN(),
                            server_default=sa.schema.DefaultClause('0'),
                            nullable=False),
                  schema='pipeline')

    # Ensure is_removed and is_delivered are set to False

    op.alter_column('file_log', 'is_archived',
                    server_default=sa.schema.DefaultClause('0'),
                    schema='pipeline')

    op.alter_column('file_log', 'is_delivered',
                    server_default=sa.schema.DefaultClause('0'),
                    schema='pipeline')

    # ### end Alembic commands ###


def downgrade():

    op.drop_column('file_log', 'to_update', schema='pipeline')

    op.alter_column('file_log', 'is_archived',
                    server_default=None,
                    schema='pipeline')

    op.alter_column('file_log', 'is_delivered',
                    server_default=None,
                    schema='pipeline')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to query the pipeline database."""
import json
import os

from sqlalchemy import and_

from roc.dingo.tools import query_db, valid_time

try:
    from poppy.core.logger import logger
    from poppy.core.db.connector import Connector
    from poppy.core.task import Task
    from poppy.core.target import PyObjectTarget, FileTarget
    from poppy.core.conf import settings
except:
    print("POPPy framework seems to not be installed properly!")

from roc.dingo.constants import PIPELINE_DATABASE, SQL_LIMIT, PIPELINE_TABLES, \
    TIME_OUTPUT_STRFORMAT, TIME_INPUT_STRFORMAT

__all__ = ["ExportToJson"]

# ________________ Global Variables _____________
# (define here the global variables)


# ________________ Class Definition __________
# (If required, define here classes)

class ExportToJson(Task):
    """
    Task to export content of the database into a JSON file.
    """
    plugin_name = 'roc.dingo'
    name = "export_to_json"

    @Connector.if_connected(PIPELINE_DATABASE)
    def run(self):

        # Initialize output dictionary
        output_json_dict = {}

        # get the pipeline connector and get the database session
        self.session = Connector.manager[PIPELINE_DATABASE].session

        # Get list of pipeline database tables to query
        table_list = self.pipeline.get('tables',
                                       default=list(PIPELINE_TABLES.keys()), args=True)

        # If --start_time keyword value is set, add an lower time limit
        start_time = valid_time(self.pipeline.get('start_time',
                                       default=[None],
                                                  args=True)[0])

        # If --end_time keyword value is set, add an upper time limit
        end_time = valid_time(self.pipeline.get('end_time',
                                     default=[None],
                                    args=True)[0])

        # If --creation-time is set to True, then
        # filtering packet_log and file_log table entries using creation time
        # instead of insertion time (not possible for invalid_packet_log)
        creation_time_flag = self.pipeline.get('creation-time',
                                       default=False, args=True)

        for current_table in table_list:

            logger.info(f'Exporting entries for {current_table}')

            # Get table model
            current_model = PIPELINE_TABLES[current_table]

            if current_table == 'packet_log':
                # Define time to use for start_time/end_time filter
                # Default is insertion time
                if creation_time_flag:
                    time = current_model.utc_time
                else:
                    time = current_model.insertion_time
            elif current_table == 'invalid_packet_log':
                time = current_model.insertion_time
            elif current_table == 'file_log':
                if creation_time_flag:
                    time = current_model.file_creation_date
                else:
                    time = current_model.file_insert_date
            else:
                logger.warning(f'Unknown table: {current_table}')
                continue

            # Add input filters
            filters = []

            if start_time:
                filters.append(time >= start_time) # greater or equal than ...

            if end_time:
                filters.append(time < end_time) # lesser than ...

            filters = and_(*filters)

            # Retrieve --limit keyword value
            limit = self.pipeline.get('limit',
                                      default=[SQL_LIMIT],
                                      args=True)[0]

            # get entries from pipeline table in ROC database
            # Order by increasing time
            entries_dict = query_db(self.session, current_model,
                                    filters=filters,
                                    limit=limit,
                                    order_by=time,
                                    to_dict='records',
                                    )
            entry_count = len(entries_dict)
            if entry_count == limit:
                logger.warning(f'Row limit has been reached {limit}')
            elif entry_count == 0:
                logger.info(f'No entry found in {current_model}')
            else:
                output_json_dict[current_table] = entries_dict

        if not output_json_dict:
            return

        # Build output file path
        output_json = self.pipeline.get('output_json', default=None)
        if output_json is None:

            database_name = 'rocdb'
            for database in self.pipeline.properties.configuration["pipeline.databases"]:
                if database['identifier'] == settings.PIPELINE_DATABASE:
                    database_name = database['login_info']['database']
                    break

            filename_field_list = [database_name,
                                   "-".join(table_list)]
            if start_time:
                filename_field_list.append(
                    start_time.strftime(TIME_OUTPUT_STRFORMAT))
            if end_time:
                filename_field_list.append(
                    end_time.strftime(TIME_OUTPUT_STRFORMAT))
            output_json = "_".join(filename_field_list) + ".json"

        output_json_path = os.path.join(self.pipeline.output, output_json)
        # Write output JSON file
        with open(output_json_path, 'w') as json_buff:
            json.dump(output_json_dict, json_buff)
        logger.info(f'{output_json_path} saved')




# ________________ Global Functions __________
# (If required, define here global functions)

